# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import tensorflow as tf
from tensorflow import keras

def get_tensor_by_shapes(shapes, batch_size = 1):
    if len(shapes) == 1:
        return tf.random.normal(shape = [batch_size] + shapes[0])
    else:
        return [tf.random.normal(shape = [batch_size] + shape) for shape in shapes]


def get_inputs_by_shapes(shapes, batch_size = 1):
    if len(shapes) == 1:
        return keras.Input(shape=shapes[0], batch_size=batch_size)
    else:
        return [keras.Input(shape=shape, batch_size=batch_size) for shape in shapes]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import logging
import jsonlines
from glob import glob
from nn_meter.predictor import latency_metrics, list_latency_predictors, load_latency_predictor
from nn_meter.utils import download_from_url, get_user_data_folder
logging = logging.getLogger("nn-Meter")


__user_dataset_folder__ = os.path.join(get_user_data_folder(), 'dataset')

def bench_dataset(url="https://github.com/microsoft/nn-Meter/releases/download/v1.0-data/datasets.zip",
                  data_folder=__user_dataset_folder__):
    if not os.path.isdir(data_folder):
        os.makedirs(data_folder)
        logging.keyinfo(f'Download from {url} ...')
        download_from_url(url, data_folder)

    datasets = glob(os.path.join(data_folder, "**.jsonl"))
    return datasets
        
if __name__ == '__main__':

    datasets = bench_dataset()
    hws = list_latency_predictors()
    
    for hw in hws:
        hw_name, hw_version = hw["name"], hw["version"]
        predictor = load_latency_predictor(hw_name, hw_version)
        for filename in datasets:
            True_lat = []
            Pred_lat = []
            index = 0
            with jsonlines.open(filename) as data_reader:
                for i, item in enumerate(data_reader):
                    graph = item["graph"]
                    pred_lat = predictor.predict(graph, model_type="nnmeter-ir")
                    real_lat = item[hw_name]
                    logging.result(f'{filename}[{i}]: predict: {pred_lat}, real: {real_lat}')
                    if real_lat != None:
                        True_lat.append(real_lat)
                        Pred_lat.append(pred_lat)
                        index += 1
            if len(True_lat) > 0:
                rmse, rmspe, error, acc5, acc10, _ = latency_metrics(Pred_lat, True_lat)
                logging.result(
                    f'{filename} on {hw_name}: rmse: {rmse}, 5%accuracy: {acc5}, 10%accuracy: {acc10}'
                )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import numpy as np
import tensorflow as tf
from tensorflow import keras
from ..interface import BaseOperator

''' 
This file contains the keras implementation of operators
'''

#---------------------- convolution layer ----------------------#

class Conv(BaseOperator):
    def get_model(self):
        cout = self.input_shape[2] if "COUT" not in self.config else self.config["COUT"]
        return keras.layers.Conv2D(
            cout,
            kernel_size=self.config["KERNEL_SIZE"],
            strides=self.config["STRIDES"],
            padding="same"
        )

    def get_output_shape(self):
        cout = self.input_shape[2] if "COUT" not in self.config else self.config["COUT"]
        output_h = (self.input_shape[0] - 1) // self.config["STRIDES"] + 1
        output_w = (self.input_shape[1] - 1) // self.config["STRIDES"] + 1
        return [output_h, output_w, cout]


class DwConv(BaseOperator):
    def get_model(self):
        return keras.layers.DepthwiseConv2D(
                kernel_size=self.config["KERNEL_SIZE"],
                strides=self.config["STRIDES"],
                padding="same"
            )

    def get_output_shape(self):
        output_h = (self.input_shape[0] - 1) // self.config["STRIDES"] + 1
        output_w = (self.input_shape[1] - 1) // self.config["STRIDES"] + 1
        return [output_h, output_w, self.input_shape[2]]


class ConvTrans(BaseOperator):
    def get_model(self):
        cout = self.input_shape[2] if "COUT" not in self.config else self.config["COUT"]
        return keras.layers.Conv2DTranspose(
                cout,
                kernel_size=self.config["KERNEL_SIZE"],
                strides=self.config["STRIDES"],
                padding="same"
            )

    def get_output_shape(self):
        cout = self.input_shape[2] if "COUT" not in self.config else self.config["COUT"]
        return [self.input_shape[0] * self.config["STRIDES"], self.input_shape[1] * self.config["STRIDES"], cout]

#------------------ normalization and pooling ------------------#

class BN(BaseOperator):
    def get_model(self):
        return keras.layers.BatchNormalization()


class GlobalAvgpool(BaseOperator):
    def get_model(self):
        return keras.layers.GlobalAveragePooling2D()

    def get_output_shape(self):
        return [self.input_shape[2]]


class MaxPool(BaseOperator):
    def get_model(self):
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        return keras.layers.MaxPool2D(
            pool_size=self.config["KERNEL_SIZE"],
            strides=self.config["POOL_STRIDES"],
            padding="same"
            )

    def get_output_shape(self):
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        output_h = (self.input_shape[0] - 1) // self.config["POOL_STRIDES"] + 1
        output_w = (self.input_shape[1] - 1) // self.config["POOL_STRIDES"] + 1
        return [output_h, output_w, self.input_shape[2]]


class AvgPool(BaseOperator):
    def get_model(self):
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        return keras.layers.AveragePooling2D(
            pool_size=self.config["KERNEL_SIZE"],
            strides=self.config["POOL_STRIDES"],
            padding="same"
            )

    def get_output_shape(self):
        if "POOL_STRIDES" not in self.config:
            self.config["POOL_STRIDES"] = self.config["STRIDES"]
        output_h = (self.input_shape[0] - 1) // self.config["POOL_STRIDES"] + 1
        output_w = (self.input_shape[1] - 1) // self.config["POOL_STRIDES"] + 1
        return [output_h, output_w, self.input_shape[2]]

#------------------------ other modules ------------------------#

class SE(BaseOperator):
    def get_model(self):
        class SE(keras.layers.Layer):
            def __init__(self, num_channels, se_ratio=0.25):
                super().__init__()
                self.pool = keras.layers.GlobalAveragePooling2D(keepdims=True)
                self.squeeze = keras.layers.Conv2D(filters=int(num_channels * se_ratio), kernel_size=1, padding='same')
                self.relu = keras.layers.ReLU()
                self.excite = keras.layers.Conv2D(filters=num_channels, kernel_size=1, padding='same')
                self.hswish = Hswish().get_model()

            def call(self, x):
                x0 = x
                x = self.pool(x)
                x = self.squeeze(x)
                x = self.relu(x)
                x = self.excite(x)
                x = self.hswish(x)
                return x * x0
        return SE(self.input_shape[-1])


class FC(BaseOperator):
    def get_model(self):
        cout = self.input_shape[-1] if "COUT" not in self.config else self.config["COUT"]
        return keras.layers.Dense(cout)

    def get_output_shape(self):
        cout = self.input_shape[-1] if "COUT" not in self.config else self.config["COUT"]
        return self.input_shape[:-1] + [cout]

#-------------------- activation function --------------------#

class Relu(BaseOperator):
    def get_model(self):
        return keras.layers.ReLU()


class Relu6(BaseOperator):
    def get_model(self):
        def func(inputs):
            return tf.nn.relu6(inputs)
        return func


class Sigmoid(BaseOperator):
    def get_model(self):
        def func(inputs):
            return tf.nn.sigmoid(inputs)
        return func


class Hswish(BaseOperator):
    def get_model(self):
        def func(inputs):
            relu6 = tf.keras.layers.ReLU(6)
            return inputs * relu6(inputs + 3.) * (1. / 6.)
        return func

#---------------------- basic operation ----------------------#

class Reshape(BaseOperator):
    def get_model(self):
        if len(self.input_shape) == 3:
            self.output_shape = [self.input_shape[2], self.input_shape[0], self.input_shape[1]]
            def func(inputs):
                return tf.reshape(inputs, [1] + self.output_shape)
        else:
            self.output_shape = [1, 2, int(self.input_shape[0] / 2)]
            def func(inputs):
                return tf.reshape(inputs, [1] + self.output_shape)
        return func

    def get_output_shape(self):
        return self.output_shape


class Add(BaseOperator):
    def get_model(self):
        return keras.layers.Add()

    def get_output_shape(self):
        if len(self.input_shape) == 2 and type(self.input_shape[0]) == list:
            output_shape = self.input_shape[0]
        else:
            output_shape = self.input_shape
        return output_shape

    def get_is_two_inputs(self):
        return True


class Concat(BaseOperator):
    def get_model(self):
        return keras.layers.Concatenate()

    def get_output_shape(self):
        if len(self.input_shape) > 1 and type(self.input_shape[0]) == list: # e.g. [[28, 28, 3], [28, 28, 5]] -> [28, 28, 8]
            output_shape = self.input_shape[0][:-1] + [sum([i[-1] for i in self.input_shape])]
        elif len(self.input_shape) == 3: # e.g. [28, 28, 4] -> [28, 28, 8]
            output_shape = self.input_shape[0:-1] + [self.input_shape[-1] * 2]
        else: # e.g. [1024] -> [2048]
            output_shape = [self.input_shape[0] * 2]
        return output_shape

    def get_is_two_inputs(self):
        return True


class Flatten(BaseOperator):
    def get_model(self):
        return keras.layers.Flatten()

    def get_output_shape(self):
        return [int(np.prod(self.input_shape))]


class Split(BaseOperator):
    def get_model(self):
        def func(inputs):
            return tf.split(inputs, num_or_size_splits=2, axis=3)
        return func

    def get_output_shape(self):
        return [self.input_shape[0], self.input_shape[1], self.input_shape[2] // 2]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .config_manager import builder_config
from .nn_meter_builder import (
    convert_models,
    profile_models,
    build_predictor_for_kernel,
    build_latency_predictor,
    build_initial_predictor_by_data,
    build_adaptive_predictor_by_data
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
CONV_TYPE = "Conv"
BN_TYPE = "BatchNormalization"
SLICE_TYPE = "Slice"
CONCAT_TYPE = "Concat"
MAXPOOL_TYPE = "MaxPool"
AVGPOOL_TYPE = "AveragePool"
RELU_TYPE = "Relu"
ADD_TYPE = "Add"
FC_TYPE = "Gemm"
RESHAPE_TYPE = "Reshape"
GAP_TYPE = "GlobalAveragePool"
CLIP_TYPE = "Clip"
MUL_TYPE = "Mul"
DIV_TYPE = "Div"
HARDSIGMOID_TYPE = "HardSigmoid"
FLATTEN_TYPE = "Flatten"
TRANSPOSE_TYPE = "Transpose"
REDUCEMEAN_TYPE = "ReduceMean"
SPLIT_TYPE = "Split"
PAD_TYPE = "Pad"



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import yaml
import pickle
import logging
from glob import glob
from nn_meter.utils import download_from_url, create_user_configs
logging = logging.getLogger("nn-Meter")


__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')


def loading_to_local(pred_info, dir):
    """ loading builtin predictors to local

    @params:

    pred_info: a dictionary containing predictor information
    dir: the local directory to store the kernel predictors and fusion rules
    """
    os.makedirs(dir, exist_ok=True)
    hardware = pred_info['name']
    ppath = os.path.join(dir, hardware)

    isdownloaded = check_predictors(ppath, pred_info["kernel_predictors"])
    if not isdownloaded:
        logging.keyinfo(f'Download from {pred_info["download"]} ...')
        download_from_url(pred_info["download"], dir)

    # load predictors
    predictors = {}
    ps = glob(os.path.join(ppath, "**.pkl"))
    for p in ps:
        pname =  os.path.basename(p).replace(".pkl", "")
        with open(p, "rb") as f:
            logging.info("load predictor %s" % p)
            model = pickle.load(f)
            predictors[pname] = model
    fusionrule = os.path.join(ppath, "fusion_rules.json")
    # logging.info(fusionrule)
    if not os.path.isfile(fusionrule):
        raise ValueError(
            "check your fusion rule path, file " + fusionrule + " does not exist！"
        )
    return predictors, fusionrule


def loading_customized_predictor(pred_info):
    """ loading customized predictor

    @params:
    pred_info: a dictionary containing predictor information
    """
    hardware = pred_info['name']
    ppath = pred_info['package_location']

    isexist = check_predictors(ppath, pred_info["kernel_predictors"])
    if not isexist:
        raise FileExistsError(f"The predictor {hardware} in {ppath} does not exist.")

    # load predictors
    predictors = {}
    ps = glob(os.path.join(ppath, "**.pkl"))
    for p in ps:
        pname =  os.path.basename(p).replace(".pkl", "")
        with open(p, "rb") as f:
            logging.info("load predictor %s" % p)
            model = pickle.load(f)
            predictors[pname] = model
    fusionrule = os.path.join(ppath, "fusion_rules.json")
    # logging.info(fusionrule)
    if not os.path.isfile(fusionrule):
        raise ValueError(
            "check your fusion rule path, file " + fusionrule + " does not exist！"
        )
    return predictors, fusionrule


def check_predictors(ppath, kernel_predictors):
    """
    @params:

    model: a pytorch/onnx/tensorflow model object or a str containing path to the model file
    """
    logging.info("checking local kernel predictors at " + ppath)
    if os.path.isdir(ppath):
        filenames = glob(os.path.join(ppath, "**.pkl"))
        # check if all the pkl files are included
        for kp in kernel_predictors:
            fullpath = os.path.join(ppath, kp + ".pkl")
            if fullpath not in filenames:
                return False
        return True
    else:
        return False


def load_config_file(fname: str, loader=None):
    """load config file from __user_config_folder__;
    if the file not located in __user_config_folder__, copy it from distribution
    """
    filepath = os.path.join(__user_config_folder__, fname)
    try:
        with open(filepath) as fp:
            if loader is None:
                return yaml.load(fp, yaml.FullLoader)
            else:
                return loader(fp)
    except FileNotFoundError:
        logging.info(f"config file {filepath} not found, created")
        create_user_configs()
        return load_config_file(fname)



import json
import numpy as np
from sklearn.metrics import mean_squared_error


def get_accuracy(y_pred, y_true, threshold = 0.01):
    a = (y_true - y_pred) / y_true
    b = np.where(abs(a) <= threshold)
    return len(b[0]) / len(y_true)


def latency_metrics(y_pred, y_true):
    rmspe = (np.sqrt(np.mean(np.square((y_true - y_pred) / y_true)))) * 100
    rmse = np.sqrt(mean_squared_error(y_pred, y_true))
    acc5 = get_accuracy(y_pred, y_true, threshold=0.05)
    acc10 = get_accuracy(y_pred, y_true, threshold=0.10)
    acc15 = get_accuracy(y_pred, y_true, threshold=0.15)
    return rmse, rmspe, rmse / np.mean(y_true), acc5, acc10, acc15


def get_conv_flop_params(hw, cin, cout, kernel_size, stride):
    params = cout * (kernel_size * kernel_size * cin + 1)
    flops = 2 * hw / stride * hw / stride * params
    return flops, params


def get_dwconv_flop_params(hw, cout, kernel_size, stride):
    params = cout * (kernel_size * kernel_size + 1)
    flops = 2 * hw / stride * hw / stride * params
    return flops, params


def get_fc_flop_params(cin, cout):
    params = (2 * cin + 1) * cout
    flops = params
    return flops, params


def get_flops_params(kernel_type, config):
    if "dwconv" in kernel_type:
        hw, cin, kernel_size, stride = config["HW"], config["CIN"], \
            config["KERNEL_SIZE"], config["STRIDES"]
        return get_dwconv_flop_params(hw, cin, kernel_size, stride)
    elif "conv" in kernel_type:
        hw, cin, cout, kernel_size, stride = config["HW"], config["CIN"], \
            config["COUT"], config["KERNEL_SIZE"], config["STRIDES"]
        return get_conv_flop_params(hw, cin, cout, kernel_size, stride)
    elif "fc" in kernel_type:
        cin, cout = config["CIN"], config["COUT"]
        return get_fc_flop_params(cin, cout)


def collect_kernel_data(kernel_data, predict_label = 'latency'):
    if isinstance(kernel_data, dict):
        return kernel_data

    config, label = kernel_data
    if isinstance(config, list):
        config = collect_data(config)
    else:
        with open(config, 'r') as fp:
            config = json.load(fp)

    if isinstance(label, list):
        label = collect_data(label)
    else:
        with open(label, 'r') as fp:
            label = json.load(fp)
    if predict_label == 'latency':
        from nn_meter.builder.backend_meta.utils import read_profiled_results
        label = read_profiled_results(label)

    for modules in config.keys():
        for model_id in config[modules].keys():
            try:
                config[modules][model_id][predict_label] = label[modules][model_id][predict_label]
            except:
                pass

    return config


def collect_data(file_list):
    file_list_copy = file_list[:]

    from ...utils import merge_info
    data = file_list_copy.pop(0)
    with open(data, 'r') as fp:
        data = json.load(fp)
    for file in file_list_copy:
        data = merge_info(new_info=file, prev_info=data)
    return data



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from functools import reduce
import copy
import math
import logging
from .protobuf_helper import ProtobufHelper as ph
logging = logging.getLogger("nn-Meter")


class ShapeInference:

    # Ops that only need to calculate the prodcast for shapes
    TF_PRODCAST_MATH_OPS = [
        "Add",
        "AddN", # AddN does not support prodcast really
        "AddV2",
        "Subtract",
        "Sub",
        "MulNoNan",
        "Multiply",
        "Mul",
        "Div",
        "DivNoNan",
        "Equal",
    ]

    # Ops that only need to propagate the shapes
    TF_PROPAGATE_MATH_OPS = [
        "Abs",
        "Acos",
        "ACosh",
        "Asin",
        "Asinh",
        "Atan",
        "Atanh",
        "Cos",
        "Cosh",
        "Exp",
        "Floor",
        "Sin",
        "Sinh",
        "Sqrt",
        "Square",

        "FusedBatchNorm",
        "FusedBatchNormV2",
        "FusedBatchNormV3",
        "BiasAdd",

        "Relu",
        "Relu6",
        "Selu",
        "LeakyReLU",
        "Elu"
        "Softmax",

        "NoOp"
    ]

    @staticmethod
    def eval_prodcast(graph, node):
        """
        Evalute the prodcast along the input nodes.
        Now we use lazy prodcast, will move to tf implement later.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        input_nodes = node["inbounds"]
        if len(input_nodes) < 2:
            logging.warn(
                "Invalid input op num for prodcast op %s" % (node["attr"]["name"])
            )
            if len(input_nodes) == 1:
                return graph[node["inbounds"][0]]["attr"]["output_shape"][0]
            else:
                return None

        target_dim = -1
        target_shape = [1]
        input_shape_list = []
        for node_name in input_nodes:
            input_shape = graph[node_name]["attr"]["output_shape"][0]
            input_shape_list.append(input_shape)
            if target_dim < len(input_shape):
                target_dim = len(input_shape)
                target_shape = input_shape
            elif target_dim == len(input_shape):
                for i in range(target_dim):
                    if target_shape[i] < input_shape[i]:
                        target_shape[i] = input_shape[i]

        return input_shape_list, [target_shape]

    @staticmethod
    def get_padding_shape(input_shape, cout, k_size, strides, padding):
        """
        Calculate the padded shape of a given tensor.

        Parameters
        ----------
        input_shape : list
            Input shape in list, a 2D or 4D list.
        cout : int
            Cout of the operation.
        k_size : list
            Kernel size of the operation, a 2D or 4D list.
        strides : list
            Strides of the operation, a 2D or 4D list.
        padding : str
            Padding type, now support SAME and VALID.
        """

        logging.info(
            "Calculating padding shape, input shape: %s, kernel size: %s, strides: %s, padding: %s."
            % (str(input_shape), str(k_size), str(strides), str(padding))
        )

        if padding == "SAME":
            outh = math.ceil(ph.get_h(input_shape) / ph.get_h(strides))
            outw = math.ceil(ph.get_w(input_shape) / ph.get_w(strides))

            padh = max(
                (outh - 1) * ph.get_h(strides)
                + ph.get_h(k_size)
                - ph.get_h(input_shape),
                0,
            )
            padw = max(
                (outw - 1) * ph.get_w(strides)
                + ph.get_w(k_size)
                - ph.get_w(input_shape),
                0,
            )

            pad_top = padh // 2
            pad_bottom = padh - pad_top
            pad_left = padw // 2
            pad_right = padw - pad_left

            pad_size = [pad_top, pad_bottom, pad_left, pad_right]
        elif padding == "VALID":
            outh = math.ceil(
                (ph.get_h(input_shape) - ph.get_h(k_size) + 1) / ph.get_h(strides)
            )
            outw = math.ceil(
                (ph.get_h(input_shape) - ph.get_h(k_size) + 1) / ph.get_w(strides)
            )

            pad_size = [0, 0, 0, 0]
        else:
            logging.error("Unexpected padding format %s." % (padding))
            return None, None

        output_shape = list(map(int, [input_shape[0], outh, outw, cout]))
        return copy.deepcopy(output_shape), copy.deepcopy(pad_size)

    @staticmethod
    def Const_get_shape(graph, node):
        """
        Get shape of a const operator.
        Take the tensor shape as the output shape.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return [], [node["attr"]["attr"]["tensor_shape"]]

    @staticmethod
    def Identity_get_shape(graph, node):
        """
        Get shape of an Identity operator.
        This is not well implemented, will move a more robust later.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return [], [graph[node["inbounds"][0]]["attr"]["output_shape"][0]]
    
    @staticmethod
    def Pad_get_shape(graph, node):
        """
        Get shape of a Pad operator.
        This is not well implemented, will move a more robust later.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        in_shape = [graph[node["inbounds"][0]]["attr"]["output_shape"][0]]
        paddings = node["attr"]["attr"]["paddings"]
        out_shape = []
        for dim in range(len(in_shape)):
            out_shape.append(in_shape[dim] + sum(paddings[dim]))
        return in_shape, out_shape
    
    @staticmethod
    def PadV2_get_shape(graph, node):
        """
        Get shape of a PadV2 operator.
        This is not well implemented, will move a more robust later.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Pad_get_shape(graph, node)

    @staticmethod
    def propagate_shape(graph, node):
        """
        For operator who does not affect the shapes.
        Just take the input shapes as output.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        logging.info("Propagate through op %s.", node["attr"]["name"])
        in_shape = [graph[node["inbounds"][0]]["attr"]["output_shape"][0]]
        return in_shape, in_shape

    @staticmethod
    def Pool_get_shape(graph, node):
        """
        Get shape of a Pooling type operation.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        if len(node["inbounds"]) != 1:
            logging.warning("Failed to get input node of %s." % (node["attr"]["name"]))
            logging.info(node)
            return

        input_shape = copy.deepcopy(
            graph[node["inbounds"][0]]["attr"]["output_shape"][0]
        )
        logging.info(
            "Get input shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], node["inbounds"][0], input_shape)
        )

        k_size = node["attr"]["attr"]["ksize"]

        if node["attr"]["attr"]["strides"][::3] != [1, 1]:
            logging.warning(
                "Invalid strides %s of node %s."
                % (str(node["attr"]["attr"]["strides"]), node["attr"]["name"])
            )
            logging.info(node)
            return

        strides = node["attr"]["attr"]["strides"]
        padding = node["attr"]["attr"]["padding"].decode("utf-8")
        logging.info(
            "Op:%s, stride:%s, padding:%s."
            % (node["attr"]["name"], str(strides), str(padding))
        )

        out_shape, padding_shape = ShapeInference.get_padding_shape(
            input_shape, input_shape[3], k_size, strides, padding
        )

        node["attr"]["attr"]["ksize"] = copy.deepcopy(
            node["attr"]["attr"]["ksize"][1:-1]
        )
        node["attr"]["attr"]["strides"] = copy.deepcopy(
            node["attr"]["attr"]["strides"][1:-1]
        )
        node["attr"]["attr"]["pads"] = copy.deepcopy(padding_shape)

        return [input_shape], [out_shape]

    @staticmethod
    def AvgPool_get_shape(graph, node):
        """
        Get shape of an AvgPool operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Pool_get_shape(graph, node)

    @staticmethod
    def AveragePooling2D_get_shape(graph, node):
        """
        Get shape of an AveragePooling2D operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Pool_get_shape(graph, node)

    @staticmethod
    def MaxPool_get_shape(graph, node):
        """
        Get shape of a MaxPool operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Pool_get_shape(graph, node)
    
    @staticmethod
    def MaxPoolV2_get_shape(graph, node):
        """
        Get shape of a MaxPool operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Pool_get_shape(graph, node)

    @staticmethod
    def MaxPooling2D_get_shape(graph, node):
        """
        Get shape of a MaxPooling2D operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Pool_get_shape(graph, node)

    @staticmethod
    def Placeholder_get_shape(graph, node):
        """
        Get shape of a Placeholder operator.
        This fetch the shape from the shape attribute of the operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return [], [node["attr"]["attr"]["shape"]]

    @staticmethod
    def Conv2D_get_shape(graph, node):
        """
        Get shape of a Conv2D operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        weight_node = ph.find_weights_root(graph, node)
        if len(weight_node) != 1:
            logging.warning("Failed to get shape of node %s." % (node["attr"]["name"]))
            logging.info(node)
            return

        input_node = [x for x in node["inbounds"] if x != weight_node]
        input_node = [x for x in input_node if graph[x]["attr"]["type"] != "Identity"]
        if len(input_node) != 1:
            logging.warning("Failed to get input node of %s." % (node["attr"]["name"]))
            logging.info(node)
            return

        input_shape = copy.deepcopy(graph[input_node[0]]["attr"]["output_shape"][0])
        logging.info(
            "Get input shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], input_node[0], input_shape)
        )

        weight_shape = graph[weight_node[0]]["attr"]["attr"]["tensor_shape"]
        if len(weight_shape) != 4:
            logging.warning(
                "Failed to parse weight shape %s of node %s."
                % (str(weight_shape), node["attr"]["name"])
            )
            logging.info(node)
            return

        logging.info(
            "Get weight shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], weight_node, weight_shape)
        )

        k_size = weight_shape[:2]
        cout = weight_shape[3]

        if node["attr"]["attr"]["strides"][::3] != [1, 1]:
            logging.warning(
                "Invalid strides %s of node %s."
                % (str(node["attr"]["attr"]["strides"]), node["attr"]["name"])
            )
            logging.info(node)
            return

        strides = node["attr"]["attr"]["strides"]
        dilation = node["attr"]["attr"]["dilations"]
        padding = node["attr"]["attr"]["padding"].decode("utf-8")
        logging.info(
            "Op:%s, stride:%s, dilation:%s, padding:%s."
            % (node["attr"]["name"], str(strides), str(dilation), str(padding))
        )

        kernel_extent_w = ph.get_w(dilation) * (ph.get_w(strides) - 1) + 1
        kernel_extent_h = ph.get_h(dilation) * (ph.get_h(strides) - 1) + 1

        out_shape, padding_shape = ShapeInference.get_padding_shape(
            input_shape, cout, [kernel_extent_w, kernel_extent_h], strides, padding
        )

        node["attr"]["attr"]["kernel_shape"] = copy.deepcopy(k_size)
        node["attr"]["attr"]["dilations"] = copy.deepcopy(
            node["attr"]["attr"]["dilations"][1:-1]
        )
        node["attr"]["attr"]["strides"] = copy.deepcopy(
            node["attr"]["attr"]["strides"][1:-1]
        )
        node["attr"]["attr"]["weight_shape"] = copy.deepcopy(weight_shape)
        node["attr"]["attr"]["pads"] = padding_shape

        return [input_shape], [out_shape]

    @staticmethod
    def DepthwiseConv2dNative_get_shape(graph, node):
        """
        Get shape of a DepthwiseConv2dNative operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        weight_node = ph.find_weights_root(graph, node)
        if len(weight_node) != 1:
            logging.warning("Failed to get shape of node %s." % (node["attr"]["name"]))
            logging.info(node)
            return

        input_node = [x for x in node["inbounds"] if x != weight_node]
        input_node = [x for x in input_node if graph[x]["attr"]["type"] != "Identity"]
        if len(input_node) != 1:
            logging.warning("Failed to get input node of %s." % (node["attr"]["name"]))
            logging.info(node)
            return

        input_shape = copy.deepcopy(graph[input_node[0]]["attr"]["output_shape"][0])
        logging.info(
            "Get input shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], input_node[0], input_shape)
        )

        weight_shape = graph[weight_node[0]]["attr"]["attr"]["tensor_shape"]
        if len(weight_shape) != 4:
            logging.warning(
                "Failed to parse weight shape %s of node %s."
                % (str(weight_shape), node["attr"]["name"])
            )
            logging.info(node)
            return

        logging.info(
            "Get weight shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], weight_node, weight_shape)
        )

        k_size = weight_shape[:2]
        cin = weight_shape[2]

        if node["attr"]["attr"]["strides"][::3] != [1, 1]:
            logging.warning(
                "Invalid strides %s of node %s."
                % (str(node["attr"]["attr"]["strides"]), node["attr"]["name"])
            )
            logging.info(node)
            return

        strides = node["attr"]["attr"]["strides"]
        dilation = node["attr"]["attr"]["dilations"]
        padding = node["attr"]["attr"]["padding"].decode("utf-8")

        logging.info(
            "Op:%s, stride:%s, dilation:%s, padding:%s."
            % (node["attr"]["name"], str(strides), str(dilation), str(padding))
        )

        kernel_extent_w = ph.get_w(dilation) * (ph.get_w(strides) - 1) + 1
        kernel_extent_h = ph.get_h(dilation) * (ph.get_h(strides) - 1) + 1

        out_shape, padding_shape = ShapeInference.get_padding_shape(
            input_shape, cin, [kernel_extent_w, kernel_extent_h], strides, padding
        )

        node["attr"]["attr"]["kernel_shape"] = copy.deepcopy(k_size)
        node["attr"]["attr"]["dilations"] = copy.deepcopy(
            node["attr"]["attr"]["dilations"][1:-1]
        )
        node["attr"]["attr"]["strides"] = copy.deepcopy(
            node["attr"]["attr"]["strides"][1:-1]
        )
        node["attr"]["attr"]["weight_shape"] = copy.deepcopy(weight_shape)
        node["attr"]["attr"]["pads"] = padding_shape

        return [input_shape], [out_shape]

    @staticmethod
    def Reduce_get_shape(graph, node):
        """
        Get shape of a Reduce operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        input_shape = graph[node["inbounds"][0]]["attr"]["output_shape"][0]
        output_shape = input_shape
        logging.info(
            "Get input shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], node["inbounds"][0], output_shape)
        )

        output_shape[1] = 0
        output_shape[2] = 0

        reduction_indices = node["attr"]["attr"]["reduction_indices"]
        logging.info("Get Reduction Indices %s.", str(reduction_indices))

        reduction_cnt = 0
        for reduction in sorted(reduction_indices):
            del output_shape[reduction - reduction_cnt]
            reduction_cnt += 1

        return [input_shape], [output_shape]

    @staticmethod
    def Mean_get_shape(graph, node):
        """
        Get shape of a Mean operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Reduce_get_shape(graph, node)

    @staticmethod
    def GlobalAveragePooling2D_get_shape(graph, node):
        """
        Get shape of a GlobalAveragePooling2D operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Reduce_get_shape(graph, node)

    @staticmethod
    def GlobalMaxPooling2D_get_shape(graph, node):
        """
        Get shape of a GlobalMaxPooling2D operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Reduce_get_shape(graph, node)

    @staticmethod
    def MatMul_get_shape(graph, node):
        """
        Get shape of a MatMul operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        weight_node = ph.find_weights_root(graph, node)
        if len(weight_node) != 1:
            logging.warning("Failed to get shape of node %s." % (node["attr"]["name"]))
            logging.info(node)
            return

        weight_shape = graph[weight_node[0]]["attr"]["attr"]["tensor_shape"]
        if len(weight_shape) != 2:
            logging.warning(
                "Failed to parse weight shape %s of node %s."
                % (str(weight_shape), node["attr"]["name"])
            )
            logging.info(node)
            return

        logging.info(
            "Get weight shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], weight_node, weight_shape)
        )

        input_node = [x for x in node["inbounds"] if x != weight_node]
        input_node = [x for x in input_node if graph[x]["attr"]["type"] != "Identity"]
        if len(input_node) != 1:
            logging.warning("Failed to get input node of %s." % (node["attr"]["name"]))
            logging.info(node)
            return

        input_shape = copy.deepcopy(graph[input_node[0]]["attr"]["output_shape"][0])
        logging.info(
            "Get input shape of %s from %s, input shape:%s."
            % (node["attr"]["name"], input_node[0], input_shape)
        )

        if weight_shape[0] != input_shape[1]:
            logging.warning(
                "Weight shape and input shape not matched for %s."
                % (node["attr"]["name"])
            )
            logging.info(node)
            return

        output_shape = copy.deepcopy(input_shape)
        output_shape[1] = weight_shape[1]

        return [input_shape], [output_shape]

    @staticmethod
    def Reshape_get_shape(graph, node):
        """
        Get shape of a Reshape operator.
        It normally should take from the shape attribution,
        but we patch it for Pack-StrideSlice-Reshape, and
        prevent the program from dynamic inference.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        if "shape" in node["attr"]["attr"].keys():
            logging.info(
                "Shape attr find in %s op, propagate with normal.", node["attr"]["name"]
            )
            input_shape = copy.deepcopy(
                graph[node["inbounds"][0]]["attr"]["output_shape"][0]
            )
            exp_output_shape = copy.deepcopy(node["attr"]["attr"]["shape"])
        else:
            logging.info(
                "Shape attr not find in %s op, try finding the shape node.",
                node["attr"]["name"],
            )
            for in_node in node["inbounds"]:
                if graph[in_node]["attr"]["type"] == "Const":
                    exp_output_shape = copy.deepcopy(
                        graph[in_node]["attr"]["constant"]
                    )
                elif graph[in_node]["attr"]["type"] == "Pack":
                    exp_output_shape = [1] + [
                        it
                        for sl in graph[in_node]["attr"]["attr"]["constant"]
                        for it in sl
                    ]
                    logging.info(
                        "Fetched expected output shape from Pack op %s"
                        % str(exp_output_shape)
                    )
                else:
                    input_shape = copy.deepcopy(
                        graph[in_node]["attr"]["output_shape"][0]
                    )

        input_elements = abs(reduce(lambda x, y: x * y, input_shape))
        exp_output_shape_elements = abs(reduce(lambda x, y: x * y, exp_output_shape))

        if input_elements != exp_output_shape_elements:
            logging.warning(
                "Input shape %s and output shape %s not matched for %s."
                % (str(input_shape), str(exp_output_shape), node["attr"]["name"])
            )

        return [input_shape], [exp_output_shape]

    @staticmethod
    def Concat_get_shape(graph, node):
        """
        Get shape of a Concat operator.
        We add up the shape through axis.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        input_shape = []
        for in_node in node["inbounds"]:
            in_shape = graph[in_node]["attr"]["output_shape"][0]
            if in_shape != []:
                input_shape.append(in_shape)
                logging.info(
                    "Get input shape of %s from %s, input shape:%s."
                    % (node["attr"]["name"], in_node, input_shape[-1])
                )
        axis = node["attr"]["attr"]["axis"][0]

        output_shape = copy.deepcopy(input_shape[0])
        for in_shape in input_shape[1:]:
            output_shape[axis] += in_shape[axis]

        return copy.deepcopy(input_shape), [output_shape]

    @staticmethod
    def Concatenate_get_shape(graph, node):
        """
        Get shape of a Concatenate operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Concat_get_shape(graph, node)

    @staticmethod
    def ConcatV2_get_shape(graph, node):
        """
        Get shape of a ConcatV2 operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        return ShapeInference.Concat_get_shape(graph, node)

    @staticmethod
    def Split_get_shape(graph, node):
        """
        Get shape of a Split operator.
        Also patched for Pack-StrideSlice-Split sequence.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        for in_node in node["inbounds"]:
            if graph[in_node]["attr"]["type"] == "Const":
                pass
            elif graph[in_node]["attr"]["type"] == "Pack":
                pass
            else:
                input_shape = copy.deepcopy(graph[in_node]["attr"]["output_shape"][0])

        split_dim = node["attr"]["attr"]["split_dim"][0]
        logging.info("Fetched Split dim for %s is %s.", node["attr"]["name"], split_dim)
        output_node_cnt = len(node["outbounds"])

        output_shape = copy.deepcopy(input_shape)
        output_shape[split_dim] = output_shape[split_dim] // output_node_cnt
        output_shape = copy.deepcopy([output_shape]) * output_node_cnt

        return [copy.deepcopy(input_shape)], output_shape

    @staticmethod
    def Transpose_get_shape(graph, node):
        """
        Get shape of a Transpose operator.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        for in_node in node["inbounds"]:
            if graph[in_node]["attr"]["type"] == "Const":
                perm = copy.deepcopy(graph[in_node]["attr"]["attr"]["constant"])
                logging.info("Fetched perm sequence from Const op %s" % str(perm))
            elif graph[in_node]["attr"]["type"] == "Pack":
                perm = [1] + [
                    it
                    for sl in graph[in_node]["attr"]["attr"]["constant"]
                    for it in sl
                ]
                logging.info("Fetched perm sequence from Pack op %s" % str(perm))
            else:
                input_shape = copy.deepcopy(graph[in_node]["attr"]["output_shape"][0])
                logging.info(
                    "Fetched input shape from %s,  %s" % (in_node, str(input_shape))
                )

        exp_output_shape = []
        for i in range(len(perm)):
            exp_output_shape.append(input_shape[perm[i]])

        return [input_shape], [exp_output_shape]

    @staticmethod
    def Pack_get_shape(graph, node):
        """
        Get shape of a Transpose operator.
        Patched for kernel detector.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        seq = ph.get_graph_seq(graph, [node["attr"]["name"]])[:5]
        for out_node in seq:
            if graph[out_node]["attr"]["type"] == "Reshape":
                if "input_shape" in graph[out_node]["attr"].keys():
                    return (
                        graph[out_node]["attr"]["input_shape"],
                        graph[out_node]["attr"]["input_shape"],
                    )
        return [[0, 0, 0, 0]], [[0, 0, 0, 0]]

    @staticmethod
    def StridedSlice_get_shape(graph, node):
        """
        Get shape of a Transpose operator.
        Patched for kernel detector.

        Parameters
        ----------
        graph : dict
            The Graph IR in dict format.
        node   : dict
            The node in Graph IR in dict format.
        """
        seq = ph.get_graph_seq(graph, [node["attr"]["name"]])[:5]
        for out_node in seq:
            if graph[out_node]["attr"]["type"] == "Reshape":
                if "input_shape" in graph[out_node]["attr"].keys():
                    return (
                        graph[out_node]["attr"]["input_shape"],
                        graph[out_node]["attr"]["input_shape"],
                    )
        return [[0, 0, 0, 0]], [[0, 0, 0, 0]]

    def __init__(self, model_graph, dynamic_fetcher):
        """
        Take the graph, and append output shape
        and input shape to the attributes of nodes.

        Parameters
        ----------
        model_graph : ModelGraph
            The ModelGraph IR class.
        """
        graph = model_graph.get_graph()
        seq = ph.get_graph_seq(graph, model_graph.get_graph_head())

        # Pass #1
        for node_name in seq:
            node_type = model_graph.get_node_type(node_name)
            node_get_shape_name = node_type + "_get_shape"

            # if node type find in supported ops, use faster static inference
            if node_type in self.TF_PRODCAST_MATH_OPS or \
                node_type in self.TF_PROPAGATE_MATH_OPS or \
                node_get_shape_name in dir(self) :

                if node_type in self.TF_PRODCAST_MATH_OPS:
                    input_shape, output_shape = ShapeInference.eval_prodcast(graph, graph[node_name])
                elif node_type in self.TF_PROPAGATE_MATH_OPS:
                    input_shape, output_shape = ShapeInference.propagate_shape(graph, graph[node_name])
                else:
                    input_shape, output_shape = eval("self." + node_get_shape_name)(
                        graph, graph[node_name]
                    )

            # fallback to dynamic inference
            # To be aware, dynamic inference does not process the shape at all, 
            # like removing weight shape from inputs. This may yield false prediction.
            else:
                logging.warn("%s is not supported by static inference yet." % model_graph.get_node_type(node_name))
                logging.warn("Failling back to dynamic fetcher, this may yield low inference speed.")
                input_shape, output_shape = dynamic_fetcher.get_shape_by_name(node_name)


            if output_shape is not None:
                graph[node_name]["attr"]["output_shape"] = copy.deepcopy(
                    output_shape
                )
            if input_shape is not None:
                graph[node_name]["attr"]["input_shape"] = copy.deepcopy(input_shape)
            logging.info(
                "Input shape of %s op is %s." % (node_name, str(input_shape))
            )
            logging.info(
                "Output shape of %s op is %s." % (node_name, str(output_shape))
            )


        # Pass #2
        # This is a patching for back-end, since backend extract shapes from
        # those two ops.
        for node_name in seq:
            if model_graph.get_node_type(node_name) in ["Pack", "StridedSlice"]:
                node_get_shape_name = model_graph.get_node_type(node_name) + "_get_shape"
                input_shape, output_shape = eval("self." + node_get_shape_name)(
                    graph, graph[node_name]
                )
                if output_shape is not None:
                    graph[node_name]["attr"]["output_shape"] = copy.deepcopy(
                        output_shape
                    )
                if input_shape is not None:
                    graph[node_name]["attr"]["input_shape"] = copy.deepcopy(input_shape)
                logging.info(
                    "Second Pass: Input shape of %s op is %s."
                    % (node_name, str(input_shape))
                )
                logging.info(
                    "Second Pass: Output shape of %s op is %s."
                    % (node_name, str(output_shape))
                )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
class MatchHelper:
    @classmethod
    def op_type_matcher(cls, node_1, node_2):
        if "type" in node_1 and "type" in node_2:
            if "_tagged" in node_1 or "_tagged" in node_2:
                return False

            if node_1["type"] == "dummy" or node_2["type"] == "dummy":
                return True
            return node_1["type"] == node_2["type"]
        else:
            return False

    @staticmethod
    def strip_useless_nodes(model_graph):
        stripped_nodes = ["Const", "Identity"]

        graph = model_graph.get_graph()
        removed_node = []
        for key, value in graph.items():
            if value["attr"]["type"] in stripped_nodes:
                removed_node.append(key)

        for key in removed_node:
            del graph[key]

        model_graph.refresh()

    @staticmethod
    def tag_matched_nodes(model_graph, matched_subgraph):
        for matched_unit in matched_subgraph:
            for node_name in matched_unit.keys():
                model_graph.add_node_attr(node_name, "_tagged", "")

    @staticmethod
    def get_untagged_nodes(model_graph):
        untagged_node = []
        for node in model_graph.get_graph().keys():
            if "_tagged" not in model_graph.get_node_attr(node)["attr"]:
                untagged_node.append(node)
        return untagged_node



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .tflite_backend import TFLiteBackend
from .tflite_profiler import TFLiteProfiler
from .cpu import TFLiteCPUBackend
from .gpu import TFLiteGPUBackend


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import numpy as np
from typing import List
from nn_meter.utils.import_package import try_import_tensorflow

class ShapeFetcher:
    def __init__(self, input_graph):
        """
        Dynamically inference the node shapes.

        Parameters
        ----------
        input_graph : graph_def
            The tensorflow input graph_def file.
        """
        self.tf = try_import_tensorflow()
        self.tf.compat.v1.disable_eager_execution()

        graph = self.tf.Graph()

        with graph.as_default():
            self.tf.import_graph_def(graph_def=input_graph, name="")
        
        self.ops = graph.get_operations()
        placeholders = list(filter(lambda op: op.type == "Placeholder", self.ops))
        assert len(placeholders) == 1
        self.graph_input_tensor = placeholders[0].outputs[0]
        graph_input_tensor_shape = self.graph_input_tensor.get_shape().as_list()
        assert graph_input_tensor_shape[1] == graph_input_tensor_shape[2]
        assert graph_input_tensor_shape[3] == 3
        self.imsize = graph_input_tensor_shape[1]
        self.graph = graph

    def get_shape_by_name(self, op_name):
        """
        Get the node output shape by its name

        Parameters
        ----------
        op_name : str
            The name of the target node.
        """
        input_tensors_to_fetch = []
        output_tensors_to_fetch = []
        for op in filter(lambda op: op.name == op_name, self.ops):
            input_tensors_to_fetch.extend(op.inputs)
            output_tensors_to_fetch.extend(op.outputs)
        
        input_shape_tensors = []
        for tensor in input_tensors_to_fetch:
            input_shape_tensors.append(self.tf.compat.v1.shape(tensor))
        
        output_shape_tensors = []
        for tensor in output_tensors_to_fetch:
            output_shape_tensors.append(self.tf.compat.v1.shape(tensor))

        
        intput_shape_results = []
        output_shape_results = []
        with self.tf.compat.v1.Session(graph=self.graph) as sess:
            fake_input = np.random.randn(1, self.imsize, self.imsize, 3)
            for shape_tensor in input_shape_tensors:
                intput_shape_results.append(sess.run(
                    shape_tensor, feed_dict={self.graph_input_tensor: fake_input}
                ).tolist())
            for shape_tensor in output_shape_tensors:
                output_shape_results.append(sess.run(
                    shape_tensor, feed_dict={self.graph_input_tensor: fake_input}
                ).tolist())

        return intput_shape_results, output_shape_results


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from sklearn.ensemble import RandomForestRegressor


__PREDICTOR_ZOO__ = {
    "conv-bn-relu": {
        "tflite_cpu": {
            "max_depth": 70,
            "n_estimators": 320,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 6,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 80,
            "n_estimators": 550,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "n_jobs": 32,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 100,
            "n_estimators": 500,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "n_jobs": 32,
            "random_state": 10,
        }
    },
    "dwconv-bn-relu": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 240,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 6,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 40,
            "n_estimators": 240,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 7,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 100,
            "n_estimators": 650,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "n_jobs": 32,
            "random_state": 10,
        }
    },
    "fc": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 70,
            "n_estimators": 330,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 4,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 70,
            "n_estimators": 330,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 4,
            "oob_score": True,
            "n_jobs": 32,
            "random_state": 10,
        }
    },
    "channelshuffle": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "se": {
        "tflite_cpu": {
            "max_depth": 20,
            "n_estimators": 290,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 110,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "maxpool": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 210,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "global-avgpool": {
        "tflite_cpu": {
            "max_depth": 70,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "hswish": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 110,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "avgpool": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 390,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "bnrelu": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 570,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "relu": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "bn": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 370,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 390,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "concat": {
        "tflite_cpu": {
            "max_depth": 100,
            "n_estimators": 690,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 100,
            "n_estimators": 690,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 5,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "addrelu": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 570,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 3,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 570,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 3,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 570,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 3,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "add": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 570,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 3,
            "oob_score": True,
            "random_state": 10,
        },
        "tflite_gpu": {
            "max_depth": 50,
            "n_estimators": 570,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 3,
            "oob_score": True,
            "random_state": 10,
        },
        "openvino_vpu": {
            "max_depth": 50,
            "n_estimators": 570,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 3,
            "oob_score": True,
            "random_state": 10,
        }
    },
    "split": {
        "tflite_cpu": {
            "max_depth": 50,
            "n_estimators": 190,
            "min_samples_leaf": 1,
            "min_samples_split": 2,
            "max_features": 2,
            "oob_score": True,
            "random_state": 10,
        }
    }
}


def init_predictor(kernel_type, backend):
    try:
        model_param = __PREDICTOR_ZOO__[kernel_type][backend]
        model = RandomForestRegressor(**model_param)
    except:
        model = RandomForestRegressor(
            max_depth = 50,
            n_estimators = 370,
            min_samples_leaf = 1,
            min_samples_split = 2,
            max_features = "auto",
            oob_score = True,
            random_state = 10,
        )
    return model



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from .utils import *
from .interface import BaseTestCase
from nn_meter.builder.backend_meta.utils import Latency


class BasicFusion(BaseTestCase):
    name = ''
    cases = {
        'ops': ['', ''],
    }
    false_case = 'ops'

    def load_config(self):
        super().load_config()
        self.eps = self.config['EMP_ALPHA']

    def test(self):
        secondary_op_lat = min(lat for op, lat in self.latency.items() if op != 'block' or op != self.false_case)
        return self.latency[self.false_case].avg - self.latency['block'].avg > self.eps * secondary_op_lat.avg

    def load_latency(self, testcase):
        self.latency['block'] = Latency(testcase['block']['latency'])

        op1, op2 = self.cases['ops']
        op1_alias, op2_alias = op1, op2

        if op1_alias == op2_alias:
            op1_alias += '_1'
            op2_alias += '_2'
        
        self.latency[op1_alias] = Latency(testcase[op1_alias]['latency'])
        self.latency[op2_alias] = Latency(testcase[op2_alias]['latency'])
        self.latency['ops'] = self.latency[op1_alias] + self.latency[op2_alias]

    def generate_testcase(self):
        testcase = {}

        op1, op2 = self.cases['ops']
        op1_alias, op2_alias = op1, op2

        if op1_alias == op2_alias:
            op1_alias += '_1'
            op2_alias += '_2'

        op1_model, op2_model, block_model, op1_shapes, op2_shapes, block_shapes = \
            generate_models_for_testcase(op1, op2, self.input_shape, self.config, self.implement)
        testcase[op1_alias] = {
            'model': op1_model,
            'shapes': op1_shapes,
        }
        testcase[op2_alias] = {
            'model': op2_model,
            'shapes': op2_shapes,
        }
        testcase['block'] = {
            'model': block_model,
            'shapes': block_shapes,
        }
        return testcase


def generate_testcases():
    testcases_list = {}
    from nn_meter.builder import builder_config
    config =  builder_config.get_module('ruletest')
    implement = builder_config.get('IMPLEMENT', 'ruletest')

    if config['BASIC_TESTCASES'] != None:
        testcases = [case.split('_') for case in config['BASIC_TESTCASES']]
        d1_required_layers = config['LAYERS_1D']
        for op1, op2 in testcases:
            class_name = f'BasicFusion_{op1}_{op2}'
            name = f'BF_{op1}_{op2}'
            cases = {
                'ops': [op1, op2],
            }
            if op1 in d1_required_layers or op2 in d1_required_layers:
                input_shape = [config['SHAPE_1D']]
            else:
                input_shape = [config['HW'], config['HW'], config['CIN']]
            bf_cls = type(class_name, (BasicFusion,), {
                'name': name,
                'cases': cases,
                'input_shape': input_shape,
                'implement': implement
            })
            testcases_list[bf_cls.name] = bf_cls
            
    if config['OTHER_TESTCASES'] != None:
        for testcase in config['OTHER_TESTCASES']:
            testcases_list[testcase] = get_special_testcases_by_name(testcase, implement=implement)

    return testcases_list



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
def get_tensor_shape(tensor):
    shape = []
    for dim in tensor.type.tensor_type.shape.dim:
        shape.append(dim.dim_value)
    if len(shape) == 4:
        shape = [shape[0], shape[2], shape[3], shape[1]]
    return shape



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
DUMMY_TYPES = [
    "Const",
    "Identity",
    "Placeholder",
]

# TODO: Refactor opset map. Should be moved to corresponding module.
OP_ALIAS = {
    # Tensorflow
    "Relu6": "relu",
    "Relu": "relu",
    "Add": "add",
    "Biasadd": "add",
    "Conv2D": "conv",
    "Reshape": "reshape",
    "FusedBatchNorm": "bn",
    "FusedBatchNormV3": "bn",
    "MatMul": "fc",
    "MaxPool": "maxpool",
    "AvgPool": "avgpool",
    "Mean": "gap",
    "Mul": "mul",
    "DepthwiseConv2dNative": "dwconv",
    "ConcatV2": "concat",
    "Split": "split",
    # ONNX
    "Conv": "conv",
    "BatchNormalization": "bn",
    "Slice": "split",
    "Concat": "concat",
    "AveragePool": "avgpool",
    "Relu": "relu",
    "Add": "add",
    "Gemm": "fc",
    "GlobalAveragePool": "gap",
    "Clip": "relu",
    "Mul": "mul",
    "Div": "div",
    "HardSigmoid": "hardsigmoid",
    "Flatten": "reshape",
    "Transpose": "transpose",
    "ReduceMean": "gap",
    "Split": "split",
    "Pad": "pad",
}



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import logging
from packaging import version
from .utils import load_config_file, loading_to_local, loading_customized_predictor
from .prediction.predict_by_kernel import nn_predict
from nn_meter.kernel_detector import KernelDetector
from nn_meter.utils import get_user_data_folder
from nn_meter.ir_converter import model_file_to_graph, model_to_graph
logging = logging.getLogger("nn-Meter")


__predictors_cfg_filename__ = 'predictors.yaml'


def list_latency_predictors():
    """return the list of latency predictors specified in nn_meter/configs/predictors.yaml
    """
    return load_config_file(__predictors_cfg_filename__)


def load_predictor_config(predictor_name: str, predictor_version: float = None):
    """
    return the information of the predictor model according to the given predictor name and version
    @params:

    predictor_name: string to specify the name of the target latency predictor. All built-in predictors can be viewed by nn_meter.list_latency_predictors() 
        or through the config file in nn_meter/configs/predictors.yaml.
    
    predictor_version: string to specify the version of the target latency predictor. If not specified (default as None), the lateast version of the 
        predictor will be loaded.
    """
    config = load_config_file(__predictors_cfg_filename__)
    preds_info = [p for p in config if p['name'] == predictor_name and (predictor_version is None or p['version'] == predictor_version)]
    n_preds = len(preds_info)
    if n_preds == 1:
        return preds_info[0]
    elif n_preds > 1:
        # find the latest version of the predictor
        latest_version, latest_version_idx = version.parse(str(preds_info[0]['version'])), 0
        for i in range(1, n_preds):
            if version.parse(str(preds_info[i]['version'])) > latest_version:
                latest_version = version.parse(str(preds_info[i]['version']))
                latest_version_idx = i
        print(f'WARNING: There are multiple version for {predictor_name}, use the latest one ({str(latest_version)})')
        return preds_info[latest_version_idx]
    else:
        raise NotImplementedError('No predictor that meets the required name and version, please try again.')


def load_latency_predictor(predictor_name: str, predictor_version: float = None):
    """ 
    return the predictor model according to the given predictor name and version
    @params:

    predictor_name: string to specify the name of the target latency predictor. All built-in predictors can be viewed by nn_meter.list_latency_predictors() 
        or through the config file in ~/.nn_meter/config/predictors.yaml.
    
    predictor_version: string to specify the version of the target latency predictor. If not specified (default as None), the lateast version of the 
        predictor will be loaded.
    """
    user_data_folder = get_user_data_folder()
    pred_info = load_predictor_config(predictor_name, predictor_version)
    if "download" in pred_info:
        kernel_predictors, fusionrule = loading_to_local(pred_info, os.path.join(user_data_folder, 'predictor'))
    else:
        kernel_predictors, fusionrule = loading_customized_predictor(pred_info)
        
    return nnMeterPredictor(kernel_predictors, fusionrule)


class nnMeterPredictor:
    def __init__(self, predictors, fusionrule):
        self.kernel_predictors = predictors
        self.fusionrule = fusionrule
        self.kd = KernelDetector(self.fusionrule)

    def predict(
        self, model, model_type, input_shape=(1, 3, 224, 224), apply_nni=False
    ):
        """
        return the predicted latency in microseconds (ms)
        @params:

        model: the model to be predicted, allowed file include
            - the path to a saved tensorflow model file (*.pb), `model_type` must be set to "pb"
            - pytorch model object (nn.Module), `model_type` must be set to "torch"
            - ONNX model object or the path to a saved ONNX model file (*.onnx), `model_type` must be set to "onnx"
            - dictionary object following nn-Meter-IR format, `model_type` must be set to "nnmeter-ir"
            - dictionary object following NNI-IR format, `model_type` must be set to "nni-ir"
            
        model_type: string to specify the type of parameter model, allowed items are ["pb", "torch", "onnx", "nnmeter-ir", "nni-ir"]
      
        input_shape: the shape of input tensor for inference (if necessary), a random tensor according to the shape will be generated and used. This parameter is only 
        accessed when model_type == 'torch'

        apply_nni: switch the torch converter used for torch model parsing. If apply_nni==True, NNI-based converter is used for torch model conversion, which requires 
            nni>=2.4 installation and should use nn interface from NNI `import nni.retiarii.nn.pytorch as nn` to define the PyTorch modules. Otherwise Onnx-based torch 
            converter is used, which requires onnx installation (well tested version is onnx==1.9.0). NNI-based converter is much faster while the conversion is unstable 
            as it could fail in some case. Onnx-based converter is much slower but stable compared to NNI-based converter. This parameter is only accessed when 
            model_type == 'torch'
        """
        logging.info("Start latency prediction ...")
        if isinstance(model, str):
            graph = model_file_to_graph(model, model_type, input_shape, apply_nni=apply_nni)
        else:
            graph = model_to_graph(model, model_type, input_shape=input_shape, apply_nni=apply_nni)
        
        # logging.info(graph)
        self.kd.load_graph(graph)

        py = nn_predict(self.kernel_predictors, self.kd.get_kernels()) # in unit of ms
        logging.info(f"Predict latency: {py} ms")
        return py



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import logging
import numpy as np
from sklearn.metrics import mean_squared_error


def get_flop(input_channel, output_channel, k, H, W, stride):
    paras = output_channel * (k * k * input_channel + 1)
    flops = 2 * H / stride * W / stride * paras
    return flops, paras


def get_conv_mem(input_channel, output_channel, k, H, W, stride):
    paras = output_channel * (k * k * input_channel + 1)
    mem = paras + output_channel * H / stride * W / stride + input_channel * H * W
    return mem


def get_depthwise_flop(input_channel, output_channel, k, H, W, stride):
    paras = output_channel * (k * k + 1)
    flops = 2 * H / stride * W / stride * paras
    return flops, paras


def get_flops_params(blocktype, hw, cin, cout, kernelsize, stride):
    if "dwconv" in blocktype:
        return get_depthwise_flop(cin, cout, kernelsize, hw, hw, stride)
    elif "conv" in blocktype:
        return get_flop(cin, cout, kernelsize, hw, hw, stride)
    elif "fc" in blocktype:
        flop = (2 * cin + 1) * cout
        return flop, flop


def get_predict_features(config):
    """
    get prediction features
    """
    mdicts = {}
    layer = 0
    for item in config:
        logging.info(item)
    for item in config:
        op = item["op"]
        if "conv" in op or "maxpool" in op or "avgpool" in op:
            cout = item["cout"]
            cin = item["cin"]
            ks = item["ks"][1]
            s = item["strides"][1] if "strides" in item else 1
            inputh = item["inputh"]
        if op in ["channelshuffle", "split"]:
            [b, inputh, inputw, cin] = item["input_tensors"][0]

        if "conv" in op:
            flops, params = get_flops_params(op, inputh, cin, cout, ks, s)
            features = [inputh, cin, cout, ks, s, flops / 2e6, params / 1e6]
        elif "fc" in op or "fc-relu" in op:
            cout = item["cout"]
            cin = item["cin"]
            flop = (2 * cin + 1) * cout
            features = [cin, cout, flop / 2e6, flop / 1e6]
        elif "pool" in op and "global" not in op:
            features = [inputh, cin, cout, ks, s]
        elif "global-pool" in op or "global-avgpool" in op or "gap" in op:
            inputh = item["inputh"] if hasattr(item, "inputh") else 1
            cin = item["cin"]
            features = [inputh, cin]
        elif "channelshuffle" in op:
            features = [inputh, cin]
        elif "split" in op:
            features = [inputh, cin]
        elif "se" in op or "SE" in op:
            inputh = item["input_tensors"][-1][-2]
            cin = item["input_tensors"][-1][-1]
            features = [inputh, cin]
        elif "concat" in op:  # maximum 4 branches
            itensors = item["input_tensors"]
            inputh = itensors[0][1]
            features = [inputh, len(itensors)]
            for it in itensors:
                co = it[-1]
                features.append(co)
            if len(features) < 6:
                features = features + [0] * (6 - len(features))
            elif len(features) > 6:
                nf = features[0:6]
                features = nf
                features[1] = 6
        elif op in ["hswish"]:
            if "inputh" in item:
                inputh = item["inputh"]
            else:
                if len(item["input_tensors"][0]) == 2:
                    inputh = item["input_tensors"][0][0]
                else:
                    inputh = item["input_tensors"][0][1]
            cin = item["cin"]
            features = [inputh, cin]
        elif op in ["bn", "relu", "bn-relu"]:
            itensors = item["input_tensors"]
            if len(itensors[0]) == 4:
                inputh = itensors[0][1]
                cin = itensors[0][3]
            else:
                inputh = itensors[0][0]
                cin = itensors[0][1]
            features = [inputh, cin]

        elif op in ["add-relu", "add"]:
            itensors = item["input_tensors"]
            inputh = itensors[0][1]
            cin1 = itensors[0][3]
            cin2 = itensors[1][3]
            features = [inputh, cin1, cin2]
        else: # indicates that there is no matching predictor for this op
            # logging.warning(f'There is no matching predictor for op {op}.')
            continue
        mdicts[layer] = {}
        mdicts[layer][op] = features
        layer += 1
    return mdicts


def read_model_latency(latency_file):
    """
    read model latency csv files. It can provide the benchmarked latency, and compare with the predicted latency
    """
    f = open(latency_file, "r")
    dicts = {}
    while True:
        line = f.readline()
        if not line:
            break
        content = line.strip().split(",")
        model = content[1]
        latency = float(content[2])
        dicts[model] = latency
    return dicts



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import logging
logging = logging.getLogger("nn-Meter")


class ProtobufHelper:
    @staticmethod
    def get_w(x):
        """
        Get width from a list.

        Parameters
        ----------
        x : list
            A 2-D or 4-D list
                represent the shape of a tensor
        """
        l = len(x)
        if l == 4:
            return x[1]
        if l == 2:
            return x[0]
        return None

    @staticmethod
    def get_h(x):
        """
        Get height from a list.

        Parameters
        ----------
        x : list
            A 2-D or 4-D list
                represent the shape of a tensor
        """
        l = len(x)
        if l == 4:
            return x[2]
        if l == 2:
            return x[1]
        return None

    @staticmethod
    def find_weights_root(graph, node):
        """
        Find the node which store the weight of the tensor.

        Parameters
        ----------
        graph : dict
            The graph IR in dict form.
        node : dict
            A single node in graph IR.
        """
        NODE_WEIGHT_LUT = {
            "Conv2D": [
                lambda x: x.replace("/Conv2D", "/weight"),
                lambda x: x.replace("/Conv2D", "/kernel"),
            ],
            "DepthwiseConv2dNative": [lambda x: x.replace("/depthwise", "/weight")],
            "BiasAdd": [
                lambda x: x.replace("/BiasAdd", "/bias"),
            ],
            "FusedBatchNorm": [
                lambda x: x.replace("/FusedBatchNormV3", "/gamma"),
                lambda x: x.replace("/FusedBatchNormV3", "/beta"),
                lambda x: x.replace("/FusedBatchNormV3", "/moving_mean"),
                lambda x: x.replace("/FusedBatchNormV3", "/moving_variance"),
            ],
            "MatMul": [
                lambda x: x.replace("/MatMul", "/weight"),
            ],
        }

        weight_name = []
        if node["attr"]["type"] in NODE_WEIGHT_LUT.keys():
            for lut_lamba in NODE_WEIGHT_LUT[node["attr"]["type"]]:
                weight_op = lut_lamba(node["attr"]["name"])
                if (
                    weight_op in graph.keys()
                    and graph[weight_op]["attr"]["type"] != "Identity"
                ):
                    logging.info(
                        "Find node %s with its weight op %s."
                        % (node["attr"]["name"], weight_op)
                    )
                    weight_name.append(weight_op)

        return weight_name

    @staticmethod
    def get_graph_seq(graph, graph_head):
        """
        Run a topological sort of the graph,
        return the sorted sequence.

        Parameters
        ----------
        graph : dict
            The graph IR in dict form.
        graph_head : str
            Start position of the sort.
        """
        seen = set()
        stack = []
        order = []
        q = [graph_head[0]]
        for head in graph_head:
            q = [head]
            while q:
                v = q.pop()
                if v not in seen:
                    seen.add(v)
                    q.extend(graph[v]["outbounds"])
                    while stack and v not in graph[stack[-1]]["outbounds"]:
                        order.append(stack.pop())
                    stack.append(v)
        return stack + order[::-1]

    @staticmethod
    def pkg42dec(x):
        """
        Convert protobuf 4-packed oct format to number.

        Parameters
        ----------
        x : list
            The 4-packed oct list.
        """
        total_byte = len(x) // 4
        assert total_byte * 4 == len(x)

        num = []
        for idx in range(total_byte):
            num.append(0)
            for i in range(4):
                num[-1] += x[idx * 4 + i] << (i * 8)
            if num[-1] == 4294967295:
                num[-1] = -1

        return num

    @staticmethod
    def get_tensor_value(x):
        """
        Get the value from a const op.

        Parameters
        ----------
        x : Protobuf.node
            The const node.
        """
        DTYPE_ENUM = {
            0: lambda x: list(map(float, x.float_val)),
            1: lambda x: list(map(float, x.float_val)),
            3: lambda x: list(map(int, x.int_val)),
        }
        data = DTYPE_ENUM[x.dtype](x)
        if len(data) == 0:
            data = ProtobufHelper.pkg42dec(x.tensor_content)
        return data



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import re
from .tflite_profiler import TFLiteProfiler
from .tflite_backend import TFLiteBackend
from ..interface import BaseParser
from nn_meter.builder.backend_meta.utils import Latency, ProfiledResults


class TFLiteCPULatencyParser(BaseParser):
    def __init__(self):
        self.nodes = []
        self.total_latency = Latency()

    def parse(self, content):
        self.nodes = self._parse_nodes(content)
        self.total_latency = self._parse_total_latency(content)
        return self

    def _parse_nodes(self, content):
        start_regex = r'[= ]*Run Order[= ]*'
        end_regex = r'[= ]*Top by Computation Time[= ]*'
        node_regex = r'\s*(\w+)\s*[\d.e-]+\s*[\d.e-]+\s*([\d.e-]+)\s*[\d.e-]+%\s*[\d.e-]+%\s*[\d.e-]+\s*1\s*(\S*)'
        flag = False

        nodes = []
        for line in content.splitlines():
            if flag:
                match = re.search(node_regex, line)
                if match:
                    nodes.append({
                        'node_type': match[1],
                        'avg': float(match[2]),
                        'name': match[3],
                    })
            if re.search(start_regex, line):
                flag = True
            if re.search(end_regex, line):
                flag = False
        
        return nodes

    def _parse_total_latency(self, content):
        total_latency_regex = r'Timings \(microseconds\): count=[\d.e-]+ first=[\d.e-]+ curr=[\d.e-]+ min=[\d.e-]+ max=[\d.e-]+ avg=([\d.e-]+) std=([\d.e-]+)'

        total_latency = Latency()
        match = re.search(total_latency_regex, content, re.MULTILINE)
        if match:
            # convert microseconds to millisecond
            total_latency = Latency(float(match[1]) / 1000, float(match[2]) / 1000)

        return total_latency
    
    @property
    def latency(self):
        return self.total_latency

    @property
    def results(self):
        results = ProfiledResults({'latency': self.latency})
        return results


class TFLiteCPUProfiler(TFLiteProfiler):
    use_gpu = False


class TFLiteCPUBackend(TFLiteBackend):
    parser_class = TFLiteCPULatencyParser
    profiler_class = TFLiteCPUProfiler



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import copy
from .constants import OP_ALIAS


def convert_nodes(graph):
    """
    Resolve inconsistency between ONNX and Tensorflow
    """
    new_graph = copy.deepcopy(graph)

    for _, node in new_graph.items():
        type = node["attr"]["type"]
        new_type = OP_ALIAS.get(type, type)
        attr = node["attr"]["attr"]

        if "kernel_shape" in attr:
            attr["ks"] = attr["kernel_shape"]
            del attr["kernel_shape"]

        if "weight_shape" in attr and attr["weight_shape"] is not None:
            attr["ks"] = attr["weight_shape"][0:2]
            del attr["weight_shape"]

        if "ksize" in attr:
            attr["ks"] = attr["ksize"]
            del attr["ksize"]

        if new_type == "split" and "axis" in attr:
            attr["split_dim"] = attr["axis"]
            del attr["axis"]

        # workaround for add, mul, div, sub with const
        if new_type in ["add", "mul", "div", "sub"] and "input_shape" in node["attr"]:
            input_shape = node["attr"]["input_shape"]
            shape = input_shape[0] if input_shape[0] else input_shape[1]
            node["attr"]["input_shape"] = [shape] * len(input_shape)

        if new_type == "conv" and "group" in attr and "input_shape" in node["attr"]:
            group = attr["group"]
            cin = node["attr"]["input_shape"][0][3]
            if group == cin:
                new_type = "dwconv"

        node["attr"]["type"] = new_type

    return new_graph



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os

from ..interface import BaseBackend
from nn_meter.utils.path import get_filename_without_ext


class OpenVINOBackend(BaseBackend):
    parser_class = None
    profiler_class = None

    def update_configs(self):
        """update the config parameters for OpenVINO platform
        """
        super().update_configs()
        self.profiler_kwargs.update({
            'venv': self.configs['OPENVINO_ENV'],
            'optimizer': self.configs['OPTIMIZER_PATH'],
            'runtime_dir': self.configs['OPENVINO_RUNTIME_DIR'],
            'serial': self.configs['DEVICE_SERIAL'],
            'data_type': self.configs['DATA_TYPE'],
        })
        self.venv = self.configs['OPENVINO_ENV']
    
    def convert_model(self, model, model_name, savedpath, input_shape=None):
        """convert the Keras model instance to frozen pb file
        """
        from .utils.converters import keras_model_to_frozenpb
        from .frozenpb_patcher import patch_frozenpb
        model_tmp_dir = os.path.join(savedpath, model_name)
        pb_path, _ = keras_model_to_frozenpb(model, model_tmp_dir, model_name, input_shape)
        patched_pb_path = patch_frozenpb(pb_path, os.path.join(self.venv, 'bin/python'))
        return patched_pb_path

    def profile(self, converted_model, metrics = ['latency'], input_shape = None):
        """convert the model to the backend platform and run the model on the backend, return required metrics 
        of the running results. We only support latency for metric by now.
        """
        self.profiler.load_graph(converted_model, self.tmp_dir)
        return self.parser.parse(self.profiler.profile(input_shape)).results.get(metrics)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import logging
from itertools import chain
from .utils import get_tensor_shape
from .constants import SLICE_TYPE
from nn_meter.utils.import_package import try_import_onnx
logging = logging.getLogger("nn-Meter")


class OnnxConverter:
    def __init__(self, model):
        onnx = try_import_onnx()
        from onnx import shape_inference
        inferred_model = shape_inference.infer_shapes(model)
        self.graph = inferred_model.graph

        self.tensors = {}
        for tensor in chain(self.graph.input, self.graph.value_info, self.graph.output):
            self.tensors[tensor.name] = {
                "shape": get_tensor_shape(tensor),
                "inputs": [],
                "outputs": [],
            }

        for node in self.graph.node:
            for input_name in node.input:
                if input_name in self.tensors:
                    self.tensors[input_name]["outputs"].append(node)
            for output_name in node.output:
                if output_name in self.tensors:
                    self.tensors[output_name]["inputs"].append(node)

    def fetch_attrs(self, node):
        from onnx import AttributeProto
        attrs = {}
        input_tensors = []
        for input_name in node.input:
            if input_name in self.tensors:
                input_tensors.append(self.tensors[input_name]["shape"])
        output_tensors = []
        for output_name in node.output:
            if output_name in self.tensors:
                output_tensors.append(self.tensors[output_name]["shape"])
        if node.op_type == SLICE_TYPE:
            for tensor_name in self._get_sibling_slice_output_tensors(node):
                output_tensors.append(self.tensors[tensor_name]["shape"])
        if (
            len(input_tensors) == 0
            or len(input_tensors[0]) <= 1
            or len(output_tensors) == 0
            or len(output_tensors[0]) <= 1
        ):
            logging.warning(f"Empty shape information with {node.name}")
            return attrs

        attrs["attr"] = {}
        attrs["type"] = node.op_type
        attrs["input_shape"] = input_tensors
        attrs["output_shape"] = output_tensors
        for attr in node.attribute:
            if attr.type == AttributeProto.FLOAT:
                attrs["attr"][attr.name] = attr.f
            elif attr.type == AttributeProto.INT:
                attrs["attr"][attr.name] = attr.i
            elif attr.type == AttributeProto.INTS:
                attrs["attr"][attr.name] = list(attr.ints)
            elif attr.type == AttributeProto.STRING:
                attrs["attr"][attr.name] = str(attr.s)
            else:
                logging.warning(f"Unsupported attributes type: {attr.type}")

        return attrs

    def convert(self):
        result = {}
        
        sliced_tensors = set()
        selected_slice = set()
        for node in self.graph.node:
            if node.op_type == SLICE_TYPE:
                tensor = node.input[0]
                if tensor in sliced_tensors:
                    continue
                else:
                    sliced_tensors.add(tensor)
                    selected_slice.add(node.name)

        for node in self.graph.node:
            outbounds = []
            inbounds = []
            if node.op_type == SLICE_TYPE and node.name not in selected_slice:
                continue
            
            for input_name in node.input:
                if input_name in self.tensors:  # remove dummy ops
                    for pred_pred in self.tensors[input_name]['inputs']:
                        inbounds.append(pred_pred.name)
            for output_name in node.output:
                if output_name in self.tensors:
                    for succ_succ in self.tensors[output_name]['outputs']:
                        outbounds.append(succ_succ.name)
                if node.op_type == SLICE_TYPE:
                    for tensor_name in self._get_sibling_slice_output_tensors(node):
                        outbounds.append(tensor_name)    
                result[node.name] = {
                    "attr": self.fetch_attrs(node),
                    "outbounds": outbounds,
                    "inbounds": inbounds,
                }

        return result

    def _get_sibling_slice_output_tensors(self, node):
        output_tensors = []
        for slice in self.tensors[node.input[0]]["outputs"]:
            if slice.name != node.name and slice.op_type == SLICE_TYPE:
                for output_name in slice.output:
                    if output_name in self.tensors:
                        output_tensors.append(output_name)

        return output_tensors



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
class UF:
    """
    UnionFind implemented with compression optimization
    """

    def __init__(self, N):
        self._parent = list(range(0, N))

    def find(self, p):
        while p != self._parent[p]:
            p = self._parent[p] = self._parent[self._parent[p]]
        return p

    def union(self, p, q):
        p = self.find(p)
        q = self.find(q)
        self._parent[q] = p

    def connected(self, p, q):
        return self.find(p) == self.find(q)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import tempfile
from ..onnx_converter import OnnxConverter
from .opset_map import nni_attr_map, nni_type_map
from nn_meter.utils.import_package import try_import_onnx, try_import_torch, try_import_onnxsim, try_import_nni


def _nchw_to_nhwc(shapes):
    return [
        [shape[0], shape[2], shape[3], shape[1]]
        if len(shape) == 4
        else shape
        for shape in shapes
    ]


class NNIIRConverter:
    def __init__(self, ir_model):
        try_import_nni()
        try:
            from nni.retiarii.converter.utils import flatten_model_graph
            self.ir_model = flatten_model_graph(ir_model)
        except:
            from nni.retiarii.converter.graph_gen import GraphConverterWithShape
            self.ir_model = ir_model.fork()
            GraphConverterWithShape().flatten(self.ir_model)

    def convert(self):
        graph = self._to_graph_layout()

        for _, node in graph.items():
            self._map_opset(node)

        self._remove_unshaped_nodes(graph)

        return graph

    def _to_graph_layout(self):
        graph = {}

        for node in self.ir_model.root_graph.hidden_nodes:
            node_dict = {
                "attr": {
                    "attr": {
                        k: v
                        for k, v in node.operation.parameters.items()
                    },
                    "input_shape": _nchw_to_nhwc(node.operation.parameters.get("input_shape")
                                                 if "input_shape" in node.operation.parameters 
                                                 else node.operation.attributes.get('input_shape')),
                    "output_shape": _nchw_to_nhwc(node.operation.parameters.get("output_shape") 
                                                 if "output_shape" in node.operation.parameters 
                                                 else node.operation.attributes.get('output_shape')),
                    "type": node.operation.type,
                },
                "inbounds": [],
                "outbounds": [],
            }

            incoming_edges = sorted(node.incoming_edges, key=lambda e: e.tail_slot or 0)
            for edge in incoming_edges:
                node_dict["inbounds"].append(edge.head.name)

            outgoing_edges = sorted(node.outgoing_edges, key=lambda e: e.head_slot or 0)
            for edge in outgoing_edges:
                node_dict["outbounds"].append(edge.tail.name)

            graph[node.name] = node_dict

        return graph

    def _map_opset(self, node):
        old_type = node["attr"]["type"]
        new_type = nni_type_map.get(old_type, old_type)

        new_attr_dict = {}
        for attr_name, attr_value in node["attr"]["attr"].items():
            new_attr_name = attr_name
            new_attr_value = attr_value
            for type, attr_map in nni_attr_map.items():
                if type == "__all__" or type == new_type:
                    if attr_name in attr_map:
                        new_attr_name, modifier = attr_map[attr_name]
                        if modifier is not None:
                            new_attr_value = modifier(attr_value)

            new_attr_dict[new_attr_name] = new_attr_value

        node["attr"]["type"] = new_type
        node["attr"]["attr"] = new_attr_dict

    def _remove_unshaped_nodes(self, graph):
        for node_name, node_dict in list(graph.items()):
            if not node_dict["attr"]["input_shape"]:
                del graph[node_name]


class NNIBasedTorchConverter(NNIIRConverter):
    def __init__(self, model, example_inputs):
        torch = try_import_torch()
        try_import_nni()
        from nni.retiarii.converter import convert_to_graph
        from nni.retiarii.converter.graph_gen import GraphConverterWithShape

        # PyTorch module to NNI IR
        script_module = torch.jit.script(model)
        converter = GraphConverterWithShape()
        ir_model = convert_to_graph(
            script_module, model, converter=converter, dummy_input=example_inputs
        )

        super().__init__(ir_model)


class OnnxBasedTorchConverter(OnnxConverter):
    def __init__(self, model, example_inputs):
        onnx = try_import_onnx()
        torch = try_import_torch()
        with tempfile.TemporaryFile() as fp:
            torch.onnx.export(model, example_inputs, fp)
            fp.seek(0)
            model = onnx.load(fp, load_external_data=False)

        # convert model
        simplify = try_import_onnxsim()
        model_simp, check = simplify(model)

        assert check, "Simplified ONNX model could not be validated"
        super().__init__(model_simp)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import sys
import yaml
import importlib

__BUILTIN_OPERATORS__ = {
    # builtin_name: module_name
    "conv": "Conv",
    "dwconv": "DwConv",
    "convtrans": "ConvTrans",
    "bn": "BN",
    "globalavgpool": "GlobalAvgpool",
    "maxpool": "MaxPool",
    "avgpool": "AvgPool",
    "se": "SE",
    "fc": "FC",
    "relu": "Relu",
    "relu6": "Relu6",
    "sigmoid": "Sigmoid",
    "hswish": "Hswish",
    "reshape": "Reshape",
    "add": "Add",
    "concat": "Concat",
    "flatten": "Flatten",
    "split": "Split"
}
__BUILTIN_TESTCASES__ = {'MON'}

__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')
__registry_cfg_filename__ = 'registry.yaml'
__REG_OPERATORS__, __REG_TESTCASES__ = {}, {}
if os.path.isfile(os.path.join(__user_config_folder__, __registry_cfg_filename__)):
    with open(os.path.join(__user_config_folder__, __registry_cfg_filename__), 'r') as fp:
        registry_modules = yaml.load(fp, yaml.FullLoader)
    if "operators" in registry_modules:
        __REG_OPERATORS__ = registry_modules["operators"]
    if "testcases" in registry_modules:
        __REG_TESTCASES__ = registry_modules["testcases"]


def get_operator_by_name(operator_name, input_shape, config = None, implement = None):
    """ get operator information by builtin name
    """
    if operator_name in __REG_OPERATORS__ and implement in __REG_OPERATORS__[operator_name]:
        operator_info = __REG_OPERATORS__[operator_name][implement]
        sys.path.append(operator_info["package_location"])
        operator_module_name = operator_info["class_name"]
        operator_module = importlib.import_module(operator_info["class_module"])

    elif operator_name in __BUILTIN_OPERATORS__:
        operator_module_name = __BUILTIN_OPERATORS__[operator_name]
        if implement == 'tensorflow':
            from nn_meter.builder.nn_modules.tf_networks import operators
        elif implement == 'torch':
            from nn_meter.builder.nn_modules.torch_networks import operators
        else:
            raise NotImplementedError('You must choose one implementation of kernel from "tensorflow" or "pytorch"')
        operator_module = operators

    else:
        raise ValueError(f"Unsupported operator name: {operator_name}. Please register the operator first.")

    operator_cls = getattr(operator_module, operator_module_name)(input_shape, config)
    operator = operator_cls.get_model()
    output_shape = operator_cls.get_output_shape()
    op_is_two_inputs = operator_cls.get_is_two_inputs()
    
    return operator, output_shape, op_is_two_inputs


def get_special_testcases_by_name(testcase, implement=None):
    if testcase in __REG_TESTCASES__ and implement in __REG_TESTCASES__[testcase]:
        testcase_info = __REG_TESTCASES__[testcase][implement]
        sys.path.append(testcase_info["package_location"])
        testcase_module_name = testcase_info["class_name"]
        testcase_module = importlib.import_module(testcase_info["class_module"])
        return getattr(testcase_module, testcase_module_name)

    elif testcase in __BUILTIN_TESTCASES__:
        assert implement != None
        if implement == 'tensorflow':
            from .build_tf_models import MultipleOutNodes
            return MultipleOutNodes
        elif implement == 'torch':
            from .build_torch_models import MultipleOutNodes
            return MultipleOutNodes
        else:
            raise NotImplementedError('You must choose one implementation of kernel from "tensorflow" or "pytorch"')

    else:
        raise ValueError(f"Unsupported operator name: {testcase}. Please register the operator first.")


def generate_models_for_testcase(op1, op2, input_shape, config, implement):
    if implement == 'tensorflow':
        from .build_tf_models import SingleOpModel, TwoOpModel
        from nn_meter.builder.nn_modules.tf_networks.utils import get_inputs_by_shapes
    elif implement == 'torch':
        from .build_torch_models import SingleOpModel, TwoOpModel
        from nn_meter.builder.nn_modules.torch_networks.utils import get_inputs_by_shapes
    else:
        raise NotImplementedError('You must choose one implementation of kernel from "tensorflow" or "pytorch"')

    layer1, op1_output_shape, op1_is_two_inputs = get_operator_by_name(op1, input_shape, config, implement)
    layer2, _, op2_is_two_inputs = get_operator_by_name(op2, op1_output_shape, config, implement)

    op1_model = SingleOpModel(layer1)
    op1_shapes = [input_shape] * (1 + op1_is_two_inputs)
    op1_model(get_inputs_by_shapes(op1_shapes))

    op2_model = SingleOpModel(layer2)
    op2_shapes = [op1_output_shape] * (1 + op2_is_two_inputs)
    op2_model(get_inputs_by_shapes(op2_shapes))

    block_model = TwoOpModel(layer1, layer2, op1_is_two_inputs, op2_is_two_inputs)
    block_shapes = [input_shape] * (1 + op1_is_two_inputs) + [op1_output_shape] * op2_is_two_inputs
    block_model(get_inputs_by_shapes(block_shapes))

    return op1_model, op2_model, block_model, op1_shapes, op2_shapes, block_shapes


def generate_single_model(op, input_shape, config, implement):
    if implement == 'tensorflow':
        from .build_tf_models import SingleOpModel
        from nn_meter.builder.nn_modules.tf_networks.utils import get_inputs_by_shapes
    elif implement == 'torch':
        from .build_torch_models import SingleOpModel
        from nn_meter.builder.nn_modules.torch_networks.utils import get_inputs_by_shapes
    else:
        raise NotImplementedError('You must choose one implementation of kernel from "tensorflow" or "pytorch"')

    layer, _, is_two_inputs = get_operator_by_name(op, input_shape, config, implement)

    model = SingleOpModel(layer)
    shapes = [input_shape] * (1 + is_two_inputs)
    model(get_inputs_by_shapes(shapes))

    return model, shapes


def save_model(model, model_path, implement):
    if implement == 'tensorflow':
        from tensorflow import keras
        from nn_meter.builder.nn_modules.tf_networks.utils import get_tensor_by_shapes
        model['model'](get_tensor_by_shapes(model['shapes']))
        keras.models.save_model(model['model'], model_path)
        return model_path

    elif implement == 'torch':
        import torch
        from nn_meter.builder.nn_modules.torch_networks.utils import get_inputs_by_shapes
        torch.onnx.export(
            model['model'],
            get_inputs_by_shapes(model['shapes']),
            model_path + '.onnx',
            input_names=['input'],
            output_names=['output'],
            verbose=False,
            export_params=True,
            opset_version=12,
            do_constant_folding=True,
        )
        return model_path + '.onnx'

    else:
        import pdb; pdb.set_trace()
        raise NotImplementedError('You must choose one implementation of kernel from "tensorflow" or "pytorch"')


def list_operators():
    return list(__BUILTIN_OPERATORS__.keys()) + \
        [f"* {key} ({', '.join(list(info.keys()))})" for key, info in __REG_OPERATORS__.items()]


def list_testcases():
    return list(__BUILTIN_TESTCASES__) + \
        [f"* {key} ({', '.join(list(info.keys()))})" for key, info in __REG_TESTCASES__.items()]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import logging
from packaging import version
logging = logging.getLogger("nn-Meter")


def try_import_onnx(require_version = ["1.10.0", "1.9.0"]):
    if isinstance(require_version, str):
        require_version = [require_version]
    try:
        import onnx
        if version.parse(onnx.__version__).release[:2] not in [version.parse(v).release[:2] for v in require_version]:
            logging.warning(f'onnx=={onnx.__version__} is not well tested now, well tested version: onnx=={", ".join(require_version)}' )
        return onnx
    except ImportError:
        logging.error(f'You have not install the onnx package, please install onnx=={require_version[0]} and try again.')


def try_import_torch(require_version = ["1.10.0", "1.9.0", "1.7.1"]):
    if isinstance(require_version, str):
        require_version = [require_version]
    try:
        import torch
        if version.parse(torch.__version__).release[:2] not in [version.parse(v).release[:2] for v in require_version]:
            logging.warning(f'torch=={torch.__version__} is not well tested now, well tested version: torch=={", ".join(require_version)}' )
        return torch
    except ImportError:
        logging.error(f'You have not install the torch package, please install torch=={require_version[0]} and try again.')


def try_import_tensorflow(require_version = ["2.7.0", "2.6.0"]):
    if isinstance(require_version, str):
        require_version = [require_version]
    try:
        import tensorflow
        if version.parse(tensorflow.__version__).release[:2] not in [version.parse(v).release[:2] for v in require_version]:
            logging.warning(f'tensorflow=={tensorflow.__version__} is not well tested now, well tested version: tensorflow=={", ".join(require_version)}' )
        return tensorflow
    except ImportError:
        logging.error(f'You have not install the tensorflow package, please install tensorflow=={require_version[0]} and try again.')


def try_import_nni(require_version = ["2.7", "2.6", "2.5", "2.4"]):
    if isinstance(require_version, str):
        require_version = [require_version]
    try:
        import nni
        if version.parse(nni.__version__).release not in [version.parse(v).release for v in require_version]:
            logging.warning(f'nni=={nni.__version__} is not well tested now, well tested version: nni=={", ".join(require_version)}' )
        return nni
    except ImportError:
        logging.error(f'You have not install the nni package, please install nni=={require_version[0]} and try again.')


def try_import_torchvision_models():
    try:
        import torchvision
        return torchvision.models
    except ImportError:
        logging.error(f'You have not install the torchvision package, please install torchvision and try again.')


def try_import_onnxsim():
    try:
        from onnxsim import simplify
        return simplify
    except ImportError:
        logging.error(f'You have not install the onnx-simplifier package, please install onnx-simplifier and try again.')


def try_import_dgl():
    try:
        import dgl
        return dgl
    except ImportError:
        logging.error(f'You have not install the dgl package, please install dgl and try again.')

    


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import json
from nn_meter.utils.graph_tool import ModelGraph
from ..utils.ir_tools import convert_nodes


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def get_fusion_unit(name):
    filename = os.path.join(BASE_DIR, f"{name}_fusionunit.json")
    with open(filename, "r") as fp:
        graph = json.load(fp)

    if not isinstance(graph, list):
        graph = [graph]

    return [ModelGraph(graph=convert_nodes(g)) for g in graph]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from nn_meter.utils.graph_tool import ModelGraph
from .utils.constants import DUMMY_TYPES
from .utils.ir_tools import convert_nodes
from .rule_reader import RuleReader
from .rule_splitter import RuleSplitter


class KernelDetector:
    def __init__(self, rule_file):
        self.reader = RuleReader(rule_file)
        self.splitter = RuleSplitter(self.reader)
        self.model_graph = None
        self.bbs = []
        self._global_index = 0

    def load_graph(self, graph):
        new_graph = convert_nodes(graph)
        self.model_graph = ModelGraph(graph=new_graph)
        self.model_graph.refresh()
        self.bbs = self.splitter.split(self.model_graph)

    def get_kernels(self):
        kernels = []
        self._global_index = 0
        self._layer_kernel_dict = {}

        for bb in self.bbs:
            kernel = self._bb_to_kernel(bb)
            self._global_index += 1
            if kernel is not None:
                kernels.append(kernel)

        self._fetch_connections(kernels)
        return kernels

    def _fetch_connections(self, kernels):
        fusion_graph = self.splitter._fusion_graph

        for kernel in kernels:
            kernel["inbounds"] = []

        for i in range(len(fusion_graph)):
            layer = fusion_graph[i]
            kernel = self._layer_kernel_dict.get(layer)

            if kernel:
                outbounds = [fusion_graph.find_root(outbound) for outbound in fusion_graph.get_outbounds(i)]
                outbounds = [self._layer_kernel_dict[outbound] for outbound in outbounds]

                for outbound in outbounds:
                    outbound["inbounds"].append(kernel["name"])

                outbounds = [outbound["name"] for outbound in outbounds]
                kernel["outbounds"] = outbounds

    def _bb_to_kernel(self, bb):
        types = [self.model_graph.get_node_type(node) for node in bb]
        # logging.info(types)
        types = [t for t in types if t and t not in DUMMY_TYPES]

        if types:
            type = "-".join(types)
            name = f"{type}#{self._global_index}"

            kernel = {
                "op": type,
                "name": name,
            }

            layer = bb[0]
            self._layer_kernel_dict[layer] = kernel
            type = types[0]
            attr = self.model_graph.get_node_attr(layer)["attr"]
            input_shape = self.model_graph.get_node_attr(layer)["input_shape"]
            output_shape = self.model_graph.get_node_attr(layer)["output_shape"]

            # Remove const from first biasadd of hswish
            if type == "hswish":
                input_shape = [input_shape[0]]
            kernel["input_tensors"] = input_shape

            if "ks" in attr:
                kernel["ks"] = attr["ks"]
            if "strides" in attr:
                kernel["strides"] = attr["strides"]
            if "split_dim" in attr:
                kernel["split_dim"] = attr["split_dim"]

            if len(input_shape) >= 1:
                if len(input_shape[0]) == 4:
                    kernel["inputh"] = input_shape[0][1]
                    kernel["inputw"] = input_shape[0][2]
                kernel["cin"] = input_shape[0][-1]

            if len(output_shape) == 1:
                kernel["cout"] = output_shape[0][-1]
            elif len(output_shape) > 1:
                kernel["output_tensors"] = output_shape

            return kernel
        else:
            return None



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import os
import yaml
import logging
import pkg_resources
from shutil import copyfile
logging = logging.getLogger("nn-Meter")


__user_config_folder__ = os.path.expanduser('~/.nn_meter/config')
__default_user_data_folder__ = os.path.expanduser('~/.nn_meter/data')

__predictors_cfg_filename__ = 'predictors.yaml'


def create_user_configs():
    """create user configs from distributed configs
    """
    os.makedirs(__user_config_folder__, exist_ok=True)
    # TODO/backlog: to handle config merging when upgrading
    for f in pkg_resources.resource_listdir(".".join(__name__.split('.')[:-2]), 'configs'):
        if f.endswith(".yaml"):
            copyfile(
                pkg_resources.resource_filename(".".join(__name__.split('.')[:-2]), f'configs/{f}'),
                os.path.join(__user_config_folder__, f))
    # make default setting yaml file
    with open(os.path.join(__user_config_folder__, 'settings.yaml'), 'w') as fp:
        yaml.dump({'data_folder': __default_user_data_folder__}, fp)


def get_user_data_folder():
    """get user data folder in settings.yaml
    """
    filepath = os.path.join(__user_config_folder__, 'settings.yaml')
    try:
        with open(filepath) as fp:
            return os.path.join(yaml.load(fp, yaml.FullLoader)['data_folder'])
    except FileNotFoundError:
        logging.info(f"setting file {filepath} not found, created")
        create_user_configs()
        return get_user_data_folder()


def change_user_data_folder(new_folder):
    """change user data folder in settings.yaml
    """
    os.makedirs(new_folder, exist_ok=True)
    with open(os.path.join(__user_config_folder__, 'settings.yaml')) as fp:
        setting = yaml.load(fp, yaml.FullLoader)
    with open(os.path.join(__user_config_folder__, 'settings.yaml'), 'w') as fp:
        setting['data_folder'] = new_folder
        yaml.dump(setting, fp)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import re
from .tflite_profiler import TFLiteProfiler
from .tflite_backend import TFLiteBackend
from ..interface import BaseParser
from nn_meter.builder.backend_meta.utils import Latency, ProfiledResults


class TFLiteGPULatencyParser(BaseParser):
    def __init__(self):
        self.kernels = []
        self.realtime = 0
        self.kernel_sum = 0
        self.block_name = ''
        self.raw_content = ''
        self.before_fused_graph = ''
        self.after_fused_graph = ''

    def parse(self, content):
        result = self._parse_time(content)
        kernel_operation_map = self._parse_kernel_name(content)
        work_size = self._parse_work_size(content)
        self.realtime, self.block_name = self._parse_block(content)
        self.kernel_sum = sum(value[0] for key, value in result.items())
        self.kernels = [{}] * len(result)
        self.before_fused_graph, self.after_fused_graph = self._parse_graph(content)
        self.comp_avg, self.comp_std = self._parse_comp_time(content)
        self.nodes = self._parse_node_cpu_time(content)
        self.errors = self._parse_error(content)
        for key, value in result.items():
            self.kernels[key] = {
                'avg': value[0],
                'std': value[1],
                'work_size': work_size[key],
                'name': kernel_operation_map[key],
            }

        self.comp_kernel_latency = sum((Latency(kernel['avg'], kernel['std']) for kernel in self.kernels if kernel['name'] != 'to/from tensor'), Latency())

        self.raw_content = content

        return self

    @staticmethod
    def resolve_name(name):
        name = name.split(' ')
        if 'linked' in name:
            ops = []
            name = [x for x in name if x != ':' and x != 'linked']
            for i in range(0, len(name), 2):
                ops.append(name[i])
            return ops
        else:
            return [name[0]]

    @property
    def latency(self):
        """
        On GPU, we currently decide to use kernel_sum instead of realtime (block) as latency
        """
        return self.comp_kernel_latency

    def _parse_kernel_name(self, content):
        kernel_name_regex = r'kernel_name\[(\d+)\]=(.*)'
        kernel_operation_map = {}

        for line in content.splitlines():
            match = re.search(kernel_name_regex, line)
            if match:
                index = int(match[1])
                kernel_operation_map[index] = match[2]

        return kernel_operation_map

    def _parse_block(self, content):
        node_regex = r'\s+\w+\s+[\d.e-]+\s+[\d.e-]+\s+([\d.e-]+)[\s\d.%]+(\S+)'

        realtime = 0
        block_name = ''
        for line in content.splitlines():
            match = re.search(node_regex, line)
            if match:
                realtime = float(match[1])
                block_name = match[2]
                break

        return realtime, block_name

    def _parse_time(self, content):
        kernel_regex = r'\w+\[(\d+)\]\w+=([\d.e-]+) \w+\[(\d+)\]\w+=([\d.e-]+) ' \
                       r'\w+\[(\d+)\]\w+=([\d.e-]+) \w+\[(\d+)\]\w+=([\d.e-]+)'
        result = {}

        for line in content.splitlines():
            match = re.search(kernel_regex, line)
            if match:
                index = int(match[1])
                avg_ms = float(match[2])
                std_ms = float(match[4])
                result[index] = (avg_ms, std_ms)

        return result

    def _parse_work_size(self, content):
        work_size_regex = r'local_work_size\[(\d+)\]=([\d,]+)'
        work_size = {}

        for line in content.splitlines():
            match = re.search(work_size_regex, line)
            if match:
                index = int(match[1])
                work_size[index] = match[2]

        return work_size

    def _parse_graph(self, content):
        before_fused_regex = r'\[Before Fused\](.*)\[end\]'
        before_fused_pattern = re.compile(before_fused_regex, re.DOTALL)

        before_fused_graph = ''
        match = before_fused_pattern.search(content)
        if match:
            before_fused_graph = match[1]

        after_fused_regex = r'\[After Fused\](.*)\[end\]'
        after_fused_pattern = re.compile(after_fused_regex, re.DOTALL)

        after_fused_graph = ''
        match = after_fused_pattern.search(content)
        if match:
            after_fused_graph = match[1]

        return before_fused_graph, after_fused_graph

    def _parse_comp_time(self, content):
        comp_time_regex = r'comp_avg_ms=([\d.e-]+) comp_std_ms=([\d.e-]+)'
        comp_avg, comp_std = 0, 0

        for line in content.splitlines():
            match = re.search(comp_time_regex, line)
            if match:
                comp_avg = float(match[1])
                comp_std = float(match[2])

        return comp_avg, comp_std

    def _parse_node_cpu_time(self, content):
        node_regex = r'(\w+)\s+[\d]+\s+([\d.e-]+)\s+[\d.%]+\s+[\d.%]+\s+[\d.e-]+\s+\d+'
        nodes = {}

        for line in content.splitlines():
            match = re.search(node_regex, line)
            if match:
                nodes[match[1]] = float(match[2])

        return nodes

    def _parse_error(self, content):
        error_regex = r'ERROR: (.*)'
        errors = []

        for line in content.splitlines():
            match = re.search(error_regex, line)
            if match:
                errors.append(match[1])

        return errors

    @property
    def results(self):
        results = ProfiledResults({'latency': self.latency})
        return results


class TFLiteGPUProfiler(TFLiteProfiler):
    use_gpu = True


class TFLiteGPUBackend(TFLiteBackend):
    parser_class = TFLiteGPULatencyParser
    profiler_class = TFLiteGPUProfiler



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from sklearn.ensemble import RandomForestRegressor


def get_model(hardware, kernel):
    model = None
    if kernel == "convbnrelu":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=70,
                n_estimators=320,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=6,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=80,
                n_estimators=550,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                n_jobs=32,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=100,
                n_estimators=500,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                n_jobs=32,
                random_state=10,
            )
    if kernel == "dwconvbnrelu":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=240,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=6,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=40,
                n_estimators=240,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=7,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=100,
                n_estimators=650,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                n_jobs=32,
                random_state=10,
            )
    if kernel == "fc":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=70,
                n_estimators=330,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=4,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=70,
                n_estimators=330,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=4,
                oob_score=True,
                n_jobs=32,
                random_state=10,
            )
    if kernel == "channelshuffle":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
    if kernel == "se":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=20,
                n_estimators=290,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=110,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
    if kernel == "maxpool":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=210,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )
    if kernel == "globalavgpool":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=70,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
    if kernel == "hswish":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=110,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )

    if kernel == "avgpool":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=390,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )
    if kernel == "bnrelu":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=570,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
    if kernel == "relu":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
    if kernel == "bn":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=370,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=390,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )
    if kernel == "concat":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=100,
                n_estimators=690,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )
        if hardware == "gpu":
            model = RandomForestRegressor(
                max_depth=100,
                n_estimators=690,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=5,
                oob_score=True,
                random_state=10,
            )

    if kernel == "addrelu":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=570,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=3,
                oob_score=True,
                random_state=10,
            )
        if hardware == "addrelu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=570,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=3,
                oob_score=True,
                random_state=10,
            )
        if hardware == "vpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=570,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=3,
                oob_score=True,
                random_state=10,
            )

    if kernel == "split":
        if hardware == "cpu":
            model = RandomForestRegressor(
                max_depth=50,
                n_estimators=190,
                min_samples_leaf=1,
                min_samples_split=2,
                max_features=2,
                oob_score=True,
                random_state=10,
            )

    return model



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import logging
import torch
import torch.nn as nn
from .operators import *
from .utils import get_inputs_by_shapes
from ..interface import BaseBlock
logging = logging.getLogger("nn-Meter")


class TorchBlock(BaseBlock):
    def __init__(self, config, batch_size = 1):
        self.config = config
        self.input_shape = [config["CIN"], config["HW"], config["HW"]]
        self.input_tensor_shape = [self.input_shape]
        self.batch_size = batch_size

    def test_block(self):
        input_data = get_inputs_by_shapes(self.input_tensor_shape, self.batch_size)
        model = self.get_model()
        model.eval()
        model(input_data)
        logging.keyinfo("Testing block is success!")

    def save_model(self, save_path):
        model = self.get_model()
        model.eval()
        torch.onnx.export(
            model,
            get_inputs_by_shapes(self.input_tensor_shape, self.batch_size),
            save_path,
            input_names=['input'],
            output_names=['output'],
            verbose=False,
            export_params=True,
            opset_version=12,
            do_constant_folding=True,
        )

    def build_model(self, ops):
        ''' convert a list of operators to torch model.
        '''
        class Model(nn.Module):
            def __init__(self, ops):
                super().__init__()
                self.layers = nn.Sequential(*ops)

            def forward(self, inputs):
                x = self.layers(inputs)
                return x

        return Model(ops)

    def get_model(self):
        raise NotImplementedError


class ConvBnRelu(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op, self.relu_op])


class ConvBnRelu6(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op, self.relu6_op])


class ConvBn(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op = bn_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op])


class ConvRelu(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.relu_op])


class ConvRelu6(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.relu6_op])


class ConvHswish(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()
        
        hswish_op = Hswish(out_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.hswish_op])


class ConvBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op = conv_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op])


class ConvBnHswish(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        hswish_op = Hswish(out_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op, self.hswish_op])


class ConvBnReluMaxPool(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        conv_op = Conv(self.input_shape, config)
        self.conv_op, out_shape = conv_op.get_model(), conv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op, out_shape = relu_op.get_model(), relu_op.get_output_shape()
        
        maxpool_op = MaxPool(out_shape, config)
        self.maxpool_op = maxpool_op.get_model()

    def get_model(self):
        return self.build_model([self.conv_op, self.bn_op, self.relu_op, self.maxpool_op])


class DwConvBn(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op = bn_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.bn_op])


class DwConvRelu(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.relu_op])


class DwConvRelu6(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.relu6_op])


class DwConvBnRelu(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.bn_op, self.relu_op])


class DwConvBnRelu6(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu6_op = Relu6(out_shape, config)
        self.relu6_op = relu6_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.bn_op, self.relu6_op])


class DwConvBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op = dwconv_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op])


class ConvBnHswish(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        dwconv_op = DwConv(self.input_shape, config)
        self.dwconv_op, out_shape = dwconv_op.get_model(), dwconv_op.get_output_shape()

        bn_op = BN(out_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        hswish_op = Hswish(out_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        return self.build_model([self.dwconv_op, self.bn_op, self.hswish_op])


class MaxPoolBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        maxpool_op = MaxPool(self.input_shape, config)
        self.maxpool_op = maxpool_op.get_model()

    def get_model(self):
        return self.build_model([self.maxpool_op])


class AvgPoolBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        avgpool_op = AvgPool(self.input_shape, config)
        self.avgpool_op = avgpool_op.get_model()

    def get_model(self):
        return self.build_model([self.avgpool_op])


class FCBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        self.config = config
        self.input_shape = [config["CIN"]]
        self.input_tensor_shape = [self.input_shape]
        self.batch_size = batch_size

        fc_op = FC(self.input_shape, config)
        self.fc_op = fc_op.get_model()

    def get_model(self):
        return self.build_model([self.fc_op])


class ConcatBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        self.config = config
        self.input_shape = [[cin, config["HW"], config["HW"]]
                       for cin in [config['CIN1'], config['CIN2'], config['CIN3'], config['CIN4']]
                       if cin != 0]
        self.input_tensor_shape = self.input_shape
        self.batch_size = batch_size
        
        concat_op = Concat(self.input_shape, config)
        self.concat_op = concat_op.get_model()

    def get_model(self):
        class Model(nn.Module):
            def __init__(self, concat_op):
                super().__init__()
                self.concat = concat_op

            def forward(self, inputs):
                return self.concat(inputs)

        return Model(self.concat_op)


class SplitBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        split_op = Split(self.input_shape, config)
        self.split_op = split_op.get_model()

    def get_model(self):
        class Model(nn.Module):
            def __init__(self, split_op):
                super().__init__()
                self.split = split_op

            def forward(self, inputs):
                return self.split(inputs)

        return Model(self.split_op)


class ChannelShuffle(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

    def get_model(self):
        class Model(nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, inputs):
                _, c, h, w = list(inputs.shape)
                x = torch.reshape(inputs, [-1, c // 2, 2, h, w])
                x = torch.transpose(x, 2, 1)
                x = torch.reshape(x, [-1, c, h, w])
                return x

        return Model()


class SEBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        se_op = SE(self.input_shape, config)
        self.se_op = se_op.get_model()

    def get_model(self):
        return self.build_model([self.se_op])


class GlobalAvgPoolBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        self.config = config
        self.input_shape = [config["HW"], config["HW"], config["CIN"]]
        self.input_tensor_shape = [self.input_shape]
        self.batch_size = batch_size

        globalavgpool_op = GlobalAvgpool(self.input_shape, config)
        self.globalavgpool_op = globalavgpool_op.get_model()

    def get_model(self):
        return self.build_model([self.globalavgpool_op])


class BnRelu(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        bn_op = BN(self.input_shape, config)
        self.bn_op, out_shape = bn_op.get_model(), bn_op.get_output_shape()
        
        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.bn_op, self.relu_op])


class BnBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        bn_op = BN(self.input_shape, config)
        self.bn_op = bn_op.get_model()

    def get_model(self):
        return self.build_model([self.bn_op])


class HswishBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        hswish_op = Hswish(self.input_shape, config)
        self.hswish_op = hswish_op.get_model()

    def get_model(self):
        return self.build_model([self.hswish_op])


class ReluBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        relu_op = Relu(self.input_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        return self.build_model([self.relu_op])


class AddRelu(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)

        add_op = Add(self.input_shape, config)
        self.add_op, out_shape = add_op.get_model(), add_op.get_output_shape()

        relu_op = Relu(out_shape, config)
        self.relu_op = relu_op.get_model()

    def get_model(self):
        class Model(nn.Module):
            def __init__(self, add_op, relu_op):
                super().__init__()
                self.add = add_op
                self.relu = relu_op

            def forward(self, inputs):
                x = self.add([inputs, inputs])
                x = self.relu(x)
                return x

        return Model(self.add_op, self.relu_op)


class AddBlock(TorchBlock):
    def __init__(self, config, batch_size = 1):
        super().__init__(config, batch_size)
        
        add_op = Add(self.input_shape, config)
        self.add_op = add_op.get_model()

    def get_model(self):
        class Model(nn.Module):
            def __init__(self, add_op):
                super().__init__()
                self.add = add_op

            def forward(self, inputs):
                return self.add([inputs, inputs])

        return Model(self.add_op)
